#ifndef BTDB_H
#define BTDB_H

/*
 * A portable database
 * - System independant, the database files should be usable on any system
 *   (the CONTENT may not be portable, though)
 * 
 */

/*  1998, 2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: btdb.h,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: btdb.h,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 * Revision 1.8  2000/07/22 12:25:03  rosmord
 * Deletion function added. Some test dones
 *
 */


#include "general.h"

#include "simplefile.h"

/* 
	 We have many problems porting this to the MAC.
	 Hence, we move from POSIX libs to ANSI for file operations
*/

/* 
	 pstring is a representation for binary data in our databases.
	 Simpler functions are provided for null terminated char * strings
*/


typedef 
struct pstring_s {
	unsigned int length;
	char * data;
} pstring;


/**
	 Internal data representation
	 linked list of data 
	 
	 BDB :: HEAD (BLOCK)*
	 HEAD :: BLOCKSIZE(1 uchar) FIRSTFREE(dbptr)
	 BLOCK :: NEXT(dbptr) PREVIOUS(dbptr) SIZE (1 uchar) RAWDATA (251 chars)

	 SIZE contains the "RAWBLOCKSIZE" value. (and must be less or equal to 256)
	 
**/

/* Size of file pointers : 4 bytes */

#define DBPTRSIZE 4

/* size of the header of a base : First free block + size of block */
#define ROOTSIZE (DBPTRSIZE + 1)

/* HEADSIZE is the size of the HEAD part in a block */
#define HEADSIZE (DBPTRSIZE*2 + 1)
/* #define RAWBLOCKSIZE 128 */
/* #define BLOCKSIZE (RAWBLOCKSIZE - HEADSIZE) */

typedef long dbptr_t;


typedef struct s_db0 {
	struct MyFile fd;										/* file pointer for the actual file */
	int flags;										/* opening flags */
	unsigned char blocksize;											/* size of raw block */
	dbptr_t firstfree;						/*  first free block*/
	int has_been_modified;				/* has the database been modified ?*/

	/* The information below should be coherent before any function use */

	int has_current;							/* is record information valid ? */
	int current_must_be_saved;		/* has current block been modified ? */	
	/* unsigned char blockhead[HEADSIZE];  not used */
	unsigned char current_block[256]; /* currently loaded block. Only valid if has_current == TRUE */
	unsigned char  position;			/* head position in current block */
	unsigned char used;						/* number of used bytes in current block */
	dbptr_t first;									/* address of first block for current record */
	dbptr_t current;								/* address of current block for current record */
	dbptr_t next, previous;				/* next and previous blocks */
} db0;

/* Structure for the tree itself : */

typedef
struct btree {
	db0 tree_nodes;
	dbptr_t root_node;						/* first node in the btree */
	db0 base;
} btdb_t;

/* 
 * tree_nodes_head is a stack of tree nodes, used to reference a
 * particular point in a given tree.  It should thus allow one to go
 * to the next node, and lots of fancy stuff like that 
*/

typedef
struct btdb_index * btdb_index_t;


/*
	Btree creation function.
  int
  btdb_open(btdb_t * tree,
					char *path,
					int flags,
					unsigned char baseblocksize,
					mode_t mode)


  -- result :
     tree is set for a database called after path.

	-- returns :
     0 if success, -1 if failure	 
*/


PUBLIC int
btdb_open(btdb_t * tree,
					char *path,
					int flags,
					unsigned char baseblocksize,
					mode_t mode);

PUBLIC void
btdb_close(btdb_t * tree);

/* Complex insertion function.*/
PUBLIC void
btdb_insert_node(btdb_t * arbre,
							 pstring * key,
							 pstring * data,
							 int (*compare)(pstring*, pstring*));
							 
							 
/* Simple insertion function */
PUBLIC void
btdb_insert_strings(btdb_t * arbre,
									char * key,
									char * data);

PUBLIC void
btdb_delete_string(btdb_t * arbre,
									 char * key);

PUBLIC int
btdb_find(btdb_t *arbre,
				pstring *key,
				pstring *data);

/*
 * btdb_apply --
 * Apply a given function f to all nodes.
 *   arbre is the btree
 *	
 *   f is a function that will be called for all nodes, 
 *   	with the following arguments :
 *      key : a pstring, the key for the current record
 *      data : a pstring, the data for the current record
 *      external data : a pointer to the external data.
 *
 * 	external_data points to some arbitrary data, which f can use
 *    instead of using some global variable. It can of course be NULL
 *
 *  keyonly is a boolean. If it is TRUE, Then the data for the records
 *      won't be retrieved, and only the key will be valid.
 *      This is interesting if you are only interested in listing the 
 *      keys, for example.
 */

PUBLIC void
btdb_apply(btdb_t * arbre,
					void (*f)(pstring *, pstring *, void *),
					void *external_data, 						
					int keyonly);

/*
 * Dump database content for debugging purposes
 */

PUBLIC void
btdb_dump(btdb_t * arbre);

/*
 * btdb_first_index --
 * sends back an index to a tree's first element.
 */

PUBLIC btdb_index_t
btdb_first_index(btdb_t * arbre);

/*
 * btdb_free_index --
 * frees an index data structure.
 */
PUBLIC void
btdb_free_index(btdb_index_t idx);

/*
 * btdb_index_to_value --
 * given a tree and an element index,
 * sends back the element's value.
 */

PUBLIC void
btdb_index_to_value(btdb_index_t idx,
										pstring *key,
										pstring *data);

/*
 * btdb_next_index --
 * given a tree and an element index,
 * sets the index to point to the next element.
 * returns 1 if success, 0 if there is no next element.
 */
PUBLIC int
btdb_next_index(btdb_index_t idx);

/*
 * btdb_previous_index --
 * given a tree and an element index,
 * sets the index to point to the previous element.
 * returns 1 if success, 0 if there is no previous element.
 */
PUBLIC int
btdb_previous_index(btdb_index_t idx);

/*
 * 
 * btdb_get_index_to --
 * gets an index to the first element greater or equal 
 * to the key.
 *
 * returns 1 if such an element exists, else 0
 */

PUBLIC int
btdb_get_index_to(btdb_t *arbre,
									btdb_index_t *idxp,
									pstring *key);

#endif
/*
 * Local Variables:
 * tab-width: 2
 * End:
 */

