/* 
	 
  sbzr, an interface function for editing bzr fonts in Tcl
						Copyright (C) 1994, 1997, 1998 Serge Rosmorduc


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

      This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

      You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

THANKS : This package had a forerunner, with whom it doesn't share a
line of code (mainly because it was easier to do everything from
scatch with SUIT) Its forerunner was written by m. Louet and Corler,
using Xlib, as a computer science project. It allowed me to edit and
correct my fonts, and I want to thank them here for it.

This version is a port of the SUIT version (called Sbzr, with capital S)
to TCL/TK.

TCL FUNCTIONS PROVIDED BY THIS PACKAGE :

bzr_loadfont => sends back a font.
bzr_savefont => saves the font to the disk
*/

/*
 * Bzr file as Tcl lists :
 * 	The Tcl representation of a Bzr file is
 *  Font ::= "{" size name bbox char* "}"
 *  char ::= code width llx lly urx ury path*
 *  path ::= "{" pathtype orient startx starty elt* "}"
 *  elt ::= "{" ("s" c1x c1y c2x c2y p2x p2y | "l" p2x p2y ) "}"
 *  bbox ::= "{" llx lly urx ury "}"
 *
 * Note that size is NOT used in the Tcl representation. That is, all
 * coordinates from the BZR file have been scaled by size when one
 * gets the Tcl list. In the reverse version (Tcl->BZR), the scaling will be done 
 * by the C function.
 *
 * More information :
 * pathtype is always "CMP". But new types will probably be interesting, for instances circles.
 *
 * orient can be "P" for positive (i.e. the path will be drawn in black
 * or "N" for negative (the path will be drawn in white). In the BZR file,
 * if the path turns in the direct (trigonometric) sense, it is a positive path, and 
 * else, a negative path.
 * 
 * The font bounding box is ignored while writing : you must provide one, but it will
 * be computed in any case. So you can consider {0 0 0 0}
 */
#define _POSIX_SOURCE 1

#include <assert.h>							/* debugging */ 
#include <limits.h>							/* guess what ? */ 
#include <math.h> 
#include <stddef.h>							/* NULL, size_t, and so on */ 
#include <stdlib.h>							/* lots of functions */ 
#include <tcl.h>
#include "bzrop.h"
#include "general.h"

#ifndef MAXDOUBLE
#define MAXDOUBLE 1.0e20
#endif 

/* 
	 in a path, pos0 is "CMP", pos 1 is the filling code, pos 2 and 3 are the first point.
	 So the first path part is at index 4
*/

#define FIRSTPATHPOS 4

#include "fordll.h"

/*
 * Declarations for functions defined in this file.
 */

EXTERN EXPORT(int,Sbzr_Init) _ANSI_ARGS_((Tcl_Interp *interp));


#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

#define MAXPTS 16

#define NEW(x) (x *)ALLOUER(sizeof(x))

typedef unsigned char Breel[3];

#define MAXREEL (255.0 + 255.0/256 + 255.0/(256*256))

/*
 * Error handler for Tcl
 *
 * We need a variable int tclcode to be defined.
 *
 */

#define TESTOK(f) if ((tclcode= f) != TCL_OK) {return tclcode;}
#define TESTOKMSG(f,m) if ((tclcode= f) != TCL_OK) {\
       Tcl_SetResult(interp, m, NULL); return tclcode;}

PRIVATE double 
read_Bdouble(FILE * f)
{
	double d;
	unsigned char c[3];
	fread(c, sizeof(char), 3, f);
	d= c[0] + (double)c[1]/ (1 << 8) + (double)c[2]/ (1 << 16);
	if (c[0] & (1 << 7))
		d= d - MAXREEL;
	return d;
}

PRIVATE void 
write_Bdouble(FILE *f, double d)
{
	unsigned char c[3];
	if (d < 0)
		d= d+ MAXREEL;
	c[0]= floor(d);
	d= (d-floor(d))*256;
	c[1]= floor(d);
	d= (d-floor(d))*256;
	c[2]= floor(d);
	fwrite(c, sizeof(char), 3,f);
}

PRIVATE double 
read_Sdouble(FILE * f)
{
	double d;
	unsigned char c[2];
	fread(c, sizeof(char), 2, f);
	d= (double)c[0]/ (1 << 8) + (double)c[1]/ (1 << 16);
	if (c[0] & (1 << 7))
		d= d - 1;
	return d;
}


/*
 * ListGetChar : l is a list. c is the first char of its i th element.
 */

PRIVATE int
ListGetChar(Tcl_Interp *interp, Tcl_Obj *l, int i, char *c)
{
	int tclcode;
	Tcl_Obj *o;
	tclcode= Tcl_ListObjIndex(interp, l, i, &o);
	if (tclcode == TCL_OK)
		{
			int l;
			char *s= Tcl_GetStringFromObj(o, &l);
			if (l > 0)
				*c= s[0];
			else
				tclcode= TCL_ERROR;
		}
	return tclcode;
}

/*
 * ListGetDouble
 * l is a list. Get its ith element as a double if possible
 */

PRIVATE int
ListGetDouble(Tcl_Interp *interp, Tcl_Obj *l, int i, double *d)
{
	int tclcode;
	Tcl_Obj *o;
	tclcode= Tcl_ListObjIndex(interp, l, i, &o);
	if (tclcode == TCL_OK)
		{
			tclcode= Tcl_GetDoubleFromObj(interp, o, d);
		}
	return tclcode;
}


/*
 * ListGetInt
 */

PRIVATE int
ListGetInt(Tcl_Interp *interp, Tcl_Obj *l, int i, int *integer)
{
	int tclcode;
	Tcl_Obj *o;
	tclcode= Tcl_ListObjIndex(interp, l, i, &o);
	if (tclcode == TCL_OK)
		{
			tclcode= Tcl_GetIntFromObj(interp, o, integer);
		}
	return tclcode;
}


/*
 * ListGetString
 */

PRIVATE int
ListGetString(Tcl_Interp *interp, Tcl_Obj *l, int i, char **s, int *length)
{
	int tclcode;
	Tcl_Obj *o;
	tclcode= Tcl_ListObjIndex(interp, l, i, &o);
	if (tclcode == TCL_OK)
		{
			*s= Tcl_GetStringFromObj(o, length);
		}
	return tclcode;
}

/*
 * computeWinding --
 * 	computes the winding of a path, either positive or negative.
 *
 *  interp : the tcl interpretor
 *  path : a Tcl list representing a path. (see path definition above)
 *  winding : the result.
 */ 

PRIVATE int
computeWinding(Tcl_Interp *interp, Tcl_Obj * path, int *winding)
{
	double area= 0.0;
	int tclcode; /* was everything ok from the Tcl point of view ? */
	int length; /* path length (in elements) */
	int i;												
	
	*winding = 1;

	tclcode= Tcl_ListObjLength(interp, path, &length);
	if (tclcode == TCL_OK && length >= FIRSTPATHPOS) {
		double px, py;

		/* read px and py */
		TESTOKMSG(ListGetDouble(interp, path, 2, &px), "double expected");
		if (tclcode == TCL_OK)
			tclcode=  ListGetDouble(interp, path, 3, &py);

		/* read the path elements */
		if (tclcode == TCL_OK) {
			for (i= FIRSTPATHPOS; i < length; i++)
				{
					Tcl_Obj *pathElt;
					
					char code;
					int tclcode;
					tclcode= Tcl_ListObjIndex(interp, path, i, &pathElt);
					tclcode= ListGetChar(interp, pathElt, 0, &code);
					switch (code)
						{
						case 'l':
							{
								double p2x, p2y;
								tclcode= ListGetDouble(interp, pathElt, 1, &p2x);
								tclcode= ListGetDouble(interp, pathElt, 2, &p2y);
								area+= (p2x -px) * (p2y +py);
								/* we should divide by 2 to get the real area, but we don't care */
								px= p2x;
								py= p2y;
							}
							break;
						case 's':
							{
								double c1x, c1y, c2x, c2y, p2x, p2y;
								double ux, uy,vx, vy ;
								double wx, wy, zx, zy;
								double x, y, t;
								tclcode= ListGetDouble(interp, pathElt, 1, &c1x);
								tclcode= ListGetDouble(interp, pathElt, 2, &c1y);
								tclcode= ListGetDouble(interp, pathElt, 3, &c2x);
								tclcode= ListGetDouble(interp, pathElt, 4, &c2y);
								tclcode= ListGetDouble(interp, pathElt, 5, &p2x);
								tclcode= ListGetDouble(interp, pathElt, 6, &p2y);
								/* Approximating the spline by just the control points 
									 is far too rough.
								*/
								ux= - px + 3*c1x - 3*c2x + p2x;
								vx= 3*c2x + 3*px -6 * c1x;
								wx= 3*c1x- 3*px;
								zx= px;
								uy= - py + 3*c1y - 3*c2y + p2y;
								vy= 3*c2y + 3*py -6 * c1y;
								wy= 3*c1y- 3*py;
								zy= py;
								for (t= 0.0; t < 1.0; t= t+1.0/MAXPTS)
									{
										x= ((ux*t+vx)*t+wx)*t+zx;
										y= ((uy*t+vy)*t+wy)*t+zy;
										area+= (x -px) * (y +py);
										px= x;
										py= y;
									}
								// Last segment in the spline :
								area += (p2x - px) *(y + p2y);
								px= p2x;
								py= p2y;
							}
							break;
						}
				}
			/* The algebraic area is  - sum(g)(y.dx) */
			if (area <= 0.0)
				*winding= 1;
			else
				*winding= -1;
		}
	} 
	return tclcode;
}

PRIVATE char 
skip_nop(FILE *f)
{
	int c;
	while ((c=getc(f)) == NO_OP);
	/* lastchar= c;*/
	return c;
}

PRIVATE int
read_font(Tcl_Interp *interp, char *name, Tcl_Obj ** res)
{
	int tclcode= TCL_OK;
	Tcl_Obj *head[2];
	Tcl_Obj *listChars= NULL;
	FILE *f;
	char c;
	double size;
	char * fontName;
	int  nameSize;

	if ((f= fopen(name,"rb")) == NULL)
		{
			Tcl_AppendResult(interp, " unknown file ", name, (char *) NULL);
			return TCL_ERROR;
		}
	
	/* Read the header : */
	/* Size */
	size= read_Bdouble(f);

	/* Font name */
	nameSize= (unsigned)getc(f);
	fontName= 
		(char *)ALLOUER((nameSize)*sizeof(char));
	fread(fontName,sizeof(char), nameSize,f);

	/* Sets the Header in Tcl Form */
	
	head[0]= Tcl_NewDoubleObj(size);
	head[1]= Tcl_NewStringObj(fontName, nameSize);
	
	/* Puts these two in the result. */
	*res= Tcl_NewListObj(2, head);
	
	/* Now, read the chars. */
	
	listChars= Tcl_NewListObj(0,0);
	while ((c=skip_nop(f)) != POST)
		{
			double width, llx, lly, urx, ury;
			Tcl_Obj *charHead[6];
			Tcl_Obj * charRep;
			int code;
			code= (unsigned)getc(f);
		
			switch(c)
				{
				case BOC_ABBREV :
					width= read_Sdouble(f) * size;
					llx= read_Sdouble(f) * size;
					lly= read_Sdouble(f) * size;
					urx= read_Sdouble(f) * size;
					ury= read_Sdouble(f) * size;
					break;
				case BOC:
					width= read_Bdouble(f) * size;
					llx= read_Bdouble(f) * size;
					lly= read_Bdouble(f) * size;
					urx= read_Bdouble(f) * size;
					ury= read_Bdouble(f) * size;
					break;
				default:
					break;
				}
			charHead[0]= Tcl_NewIntObj(code);
			charHead[1]= Tcl_NewDoubleObj(width);
			charHead[2]= Tcl_NewDoubleObj(llx);
			charHead[3]= Tcl_NewDoubleObj(lly);
			charHead[4]= Tcl_NewDoubleObj(urx);
			charHead[5]= Tcl_NewDoubleObj(ury);
			charRep= Tcl_NewListObj(6, charHead);
		
			/* paths */
			c= skip_nop(f);
			while ( c != EOC)
				{
					double px, py;
					char * type;
					/* A count of significant path elements.
						 A path isn't interesting if it doesn't contain
						 at least one spline or two line segments. 
					*/
					int significantElements= 0;

					Tcl_Obj *pathHead[4];
					Tcl_Obj * path;
		
					px= read_Bdouble(f) * size;
					py= read_Bdouble(f) * size;
			
					switch (c) {
					case START_CURVE:
						type= "T";
						break;
					case START_NOFIL:
						type= "S";
						break;
					case START_PATH:
						type= "P";
						break;
					}

					pathHead[0]= Tcl_NewStringObj("CMP",3);
					pathHead[1]= Tcl_NewStringObj(type,1);
					pathHead[2]= Tcl_NewDoubleObj(px);
					pathHead[3]= Tcl_NewDoubleObj(py);
					path= Tcl_NewListObj(4, pathHead);
			
					while ((c= skip_nop(f)) != START_PATH 
								 && c != START_CURVE
								 && c != START_NOFIL
								 && c != EOC)
						{
							Tcl_Obj *pathElt;

							switch(c)
								{
								case LINE:
									{
										Tcl_Obj *t[3];
										t[0]= Tcl_NewStringObj("l",1);
										t[1]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										t[2]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										pathElt= Tcl_NewListObj(3, t);
										significantElements++;
										break;
									}
								case LINE_ABBREV:
									{
										Tcl_Obj *t[3];
										t[0]= Tcl_NewStringObj("l",1);
										t[1]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										t[2]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										pathElt= Tcl_NewListObj(3, t);
										significantElements++;
										break;
									}
								case SPLINE:
									{
										Tcl_Obj *t[7];
										t[0]= Tcl_NewStringObj("s",1);
										/* c1 */
										t[1]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										t[2]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										/* c2 */
										t[3]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										t[4]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										/* p2 */
										t[5]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										t[6]= Tcl_NewDoubleObj(read_Bdouble(f) * size);
										pathElt= Tcl_NewListObj(7, t);
										significantElements+= 2;
										break;
									}
								case SPLINE_ABBREV:
									{
										Tcl_Obj *t[7];
										t[0]= Tcl_NewStringObj("s",1);
										/* c1 */
										t[1]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										t[2]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										/* c2 */
										t[3]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										t[4]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										/* p2 */
										t[5]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										t[6]= Tcl_NewDoubleObj(read_Sdouble(f) * size);
										pathElt= Tcl_NewListObj(7, t);
										significantElements+= 2;
										break;
									}
								}
							/* Add the path element to the current path : */
							Tcl_ListObjAppendElement(interp, path, pathElt);
						}
					if (type[0] == 'P')
						{
							int winding;
							/* Compute the path winding. */
							tclcode= computeWinding(interp, path, &winding);
							if (winding < 0)
								{
									Tcl_Obj *newType[1];
									newType[0]= Tcl_NewStringObj("N",1);
									tclcode= Tcl_ListObjReplace(interp, path, 1, 1, 1, newType);
								}
						}
					/* if the path is reduced to one line segment, or worse, one point, 
						 we don't need it. */
					/* else add it to the charRep */
					if (significantElements <= 1)
						{
							/* NOTE : Tcl_DecrRefCount is a MACRO. The accolade {} around it
								 are mandatory, because if they are not here, the following else 
								 will be misunderstood.
							*/

							Tcl_DecrRefCount(path);
						}
					else
						{
							Tcl_ListObjAppendElement(interp, charRep, path);
						}
				}
			/* Add the char to the font */
			Tcl_ListObjAppendElement(interp, listChars, charRep);
		} 
	/*
	 * Now, c == POST. We read the bounding box data : 
	 */
	if (c == POST) { /* Normally always TRUE here*/
		int i;
		Tcl_Obj *t[4];
		printf("reading bb\n");
		for(i=0; i< 4; i++) {
			double x;
			x= read_Bdouble(f) * size;
			printf("x=%f ", x);
			t[i]= Tcl_NewDoubleObj(x);
		}
		printf("\n");
		/*
		 * Now we add the bounding box as a list to the result
		 */
		Tcl_ListObjAppendElement(interp, *res, Tcl_NewListObj(4, t));
	}
	/*
	 * and then all the characters are added to the result :
	 */
	Tcl_ListObjAppendList(interp, *res, listChars);
	fclose(f);
	return tclcode;
}

/*
 * write_BZRFont --
 * 	takes a Tcl list, and write the corresponding BZR file.
 */

PRIVATE int
write_fonte(Tcl_Interp *interp, Tcl_Obj* fontList, char *fname)
{
	int tclcode;
	FILE *f;
	double size;
	double llx= MAXDOUBLE;
	double lly= MAXDOUBLE;
	double urx= -MAXDOUBLE;
	double ury= -MAXDOUBLE;
	unsigned int nbrchar=0;
	unsigned int where= 0;
	unsigned int postptr;
	unsigned int wherelettres[256];
	unsigned int used[256];				/* which letter positions are used */
	int i;

	tclcode= TCL_ERROR;

	for (i= 0; i < 256; i++)
		{
			used[i]= 0;
		}
	
	if ((f= fopen(fname, "wb")))
	{
		unsigned int i;
		char * fontName;
		int fontNameLength;
		int length;

		/* The size */
		tclcode= ListGetDouble(interp, fontList, 0, &size);
		write_Bdouble(f, size);

		/* The name : length, then chars. */
		
		ListGetString(interp, fontList, 1, &fontName, &fontNameLength);
		
		putc(fontNameLength,f);
		fputs(fontName,f);
		where= 3 + 1+ fontNameLength;

		tclcode= Tcl_ListObjLength(interp, fontList, &length);
		
		/*
		 * We start writing chars now. The first char index is i=3.
		 * index 2 is the bounding box, which we ignore.
		 *
		 */
		
		for (i= 3; i<length; i++)
		{
			int code;
			double width;
			double cllx, clly, curx, cury; /* char bounding box */
			Tcl_Obj *letter;
			int letterLength;
			int j;

			
			tclcode= Tcl_ListObjIndex(interp, fontList, i, &letter);
			
			tclcode= ListGetInt(interp, letter, 0, &code);
			tclcode= ListGetDouble(interp, letter, 1, &width);
			tclcode= ListGetDouble(interp, letter, 2, &cllx);
			tclcode= ListGetDouble(interp, letter, 3, &clly);
			tclcode= ListGetDouble(interp, letter, 4, &curx);
			tclcode= ListGetDouble(interp, letter, 5, &cury);

			nbrchar++;
			wherelettres[code]= where;
			used[code]= TRUE;
			
			putc(BOC,f); where++;
			putc(code,f); where++;
			
			write_Bdouble(f, width/size);
			write_Bdouble(f, cllx/size);
			write_Bdouble(f, clly/size);
			write_Bdouble(f, curx/size);
			write_Bdouble(f, cury/size);
			where+= 5*3;
			/*
			 * Fix the font's global bounding box.
			 */

			if (llx > cllx)
				llx=  cllx;
			if (lly > clly)
				lly=  clly;
			if (urx < curx)
				urx=  curx;
			if (ury < cury)
				ury=  cury;
			
			tclcode= Tcl_ListObjLength(interp, letter, &letterLength);

			/* Now, we loop on the paths */
			for(j= 6; j < letterLength; j++)
				{
					int pathLength;
					Tcl_Obj *path;
					int k;
					char pathCode;
					double px, py;
					int winding;
					int reverse;
					
					tclcode= Tcl_ListObjIndex(interp, letter, j, &path);

					
					/* The first value is currently always "CMP" */
					/* Path Code */
					tclcode= ListGetChar(interp, path, 1, &pathCode);
					/* px and py */
					tclcode= ListGetDouble(interp, path, 2, &px);
					tclcode= ListGetDouble(interp, path, 3, &py);
					
					tclcode= Tcl_ListObjLength(interp, path, &pathLength);

					switch (pathCode)
						{
						case 'P':
						case 'N':
							putc(START_PATH,f); where++;
							break;
						case 'S':
							putc(START_NOFIL,f); where++;
							break;
						case 'T':
							putc(START_CURVE,f); where++;
							break;
						}
					
					/*
					 * The first point will always be the same, even if reversing the path. 
					 */

					write_Bdouble(f, px/size);				
					write_Bdouble(f, py/size);
					where+= 2*3;
					

					/* Now, we want to get sure that the path is written 
					 * in the proper order.
					 */
					reverse= FALSE;
					if (pathCode == 'N' || pathCode == 'P')
						{
							tclcode= computeWinding(interp, path, &winding);
							if ((pathCode == 'N' && winding >= 0 )
									|| (pathCode == 'P' && winding < 0))
								reverse= TRUE;
						}
					
					if (reverse) 
						{
							double firstx, firsty;
							firstx= px;
							firsty= py;
							
							/* start writing the END of the path */
							/* Loop on path elements. */
							
							/*
							 * the problem is that if we have for instance a path p0 {s c1 c2 p1} {c3 c4 p2} {t p0}
							 * we need to look at the previous element when printing the current one. That is, the first
							 * element will be {t p2} and not {t p0}, which means we need to look at {c3 c4 p2}
							 *
							 */
							
							for(k= pathLength - 1; k >= FIRSTPATHPOS; k--)
								{
									Tcl_Obj *pathElt;
									double p2x, p2y;
									
									char eltCode;

									tclcode= Tcl_ListObjIndex(interp, path, k, &pathElt);
									/* get the point from the "previous" element */
									if (k > 0)
										{
											char prevCode;
											Tcl_Obj *prevPathElt;

											tclcode= Tcl_ListObjIndex(interp, path, k - 1, &prevPathElt);
											tclcode= ListGetChar(interp, prevPathElt, 0, &prevCode);

											switch (prevCode) {
											case 'l':
												tclcode= ListGetDouble(interp, prevPathElt, 1, &p2x);
												tclcode= ListGetDouble(interp, prevPathElt, 2, &p2y);
												break;
											case 's':
												tclcode= ListGetDouble(interp, prevPathElt, 5, &p2x);
												tclcode= ListGetDouble(interp, prevPathElt, 6, &p2y);
												break;
											}
										}
									else
										{
											p2x= firstx;
											p2y= firsty;
										}

									tclcode= ListGetChar(interp, pathElt, 0, &eltCode);
									switch (eltCode)
										{
										case 'l':
											{
												putc(LINE,f); where++;
												write_Bdouble(f, p2x/size);
												write_Bdouble(f, p2y/size);
												where+= 2*3;
											}
											break;
										case 's':
											{
												double c1x, c1y, c2x, c2y;
												tclcode= ListGetDouble(interp, pathElt, 1, &c1x);
												tclcode= ListGetDouble(interp, pathElt, 2, &c1y);
												tclcode= ListGetDouble(interp, pathElt, 3, &c2x);
												tclcode= ListGetDouble(interp, pathElt, 4, &c2y);
												
												putc(SPLINE,f); where++;

												write_Bdouble(f, c2x/size);
												write_Bdouble(f, c2y/size);
												write_Bdouble(f, c1x/size);
												write_Bdouble(f, c1y/size);
												write_Bdouble(f, p2x/size);
												write_Bdouble(f, p2y/size);
												where+= 6*3;
											}
											break;
										}
								}

						}
					else
						{
							/* start writing the beginning of the path.
							 * definitly simpler.  */
							
							/* Loop on path elements. */
							for(k= FIRSTPATHPOS; k < pathLength; k++)
								{
									Tcl_Obj *pathElt;
									char eltCode;
									
									tclcode= Tcl_ListObjIndex(interp, path, k, &pathElt);
									tclcode= ListGetChar(interp, pathElt, 0, &eltCode);
									switch (eltCode)
										{
										case 'l':
											{
												double p2x, p2y;
												
												tclcode= ListGetDouble(interp, pathElt, 1, &p2x);
												tclcode= ListGetDouble(interp, pathElt, 2, &p2y);

												putc(LINE,f); where++;
												write_Bdouble(f, p2x/size);
												write_Bdouble(f, p2y/size);
												where+= 2*3;
											}
											break;
										case 's':
											{
												double c1x, c1y, c2x, c2y, p2x, p2y;
												tclcode= ListGetDouble(interp, pathElt, 1, &c1x);
												tclcode= ListGetDouble(interp, pathElt, 2, &c1y);
												tclcode= ListGetDouble(interp, pathElt, 3, &c2x);
												tclcode= ListGetDouble(interp, pathElt, 4, &c2y);
												tclcode= ListGetDouble(interp, pathElt, 5, &p2x);
												tclcode= ListGetDouble(interp, pathElt, 6, &p2y);
												
												putc(SPLINE,f); where++;
												
												write_Bdouble(f, c1x/size);
												write_Bdouble(f, c1y/size);
												write_Bdouble(f, c2x/size);
												write_Bdouble(f, c2y/size);
												write_Bdouble(f, p2x/size);
												write_Bdouble(f, p2y/size);
												where+= 6*3;
											}
											break;
										}
								}
						}

				}
			putc(EOC,f); where++;
		}
		
		postptr= where;
		putc(POST,f); where++;
		write_Bdouble(f, llx/size);
		write_Bdouble(f, lly/size);
		write_Bdouble(f, urx/size);		
		write_Bdouble(f, ury/size);
		/* printf("bounding box %f %f %f %f\n", llx/size, lly/size, urx/size, ury/size); */
		where+= 4*3;
		for (i=0; i<256; i++)
		{
			if (used[i])
			{
				unsigned int ou= wherelettres[i];
				putc(CHAR_LOC,f); where++;
				putc(i,f); where++;
				putc(ou >> 8*3,f); where++;
				putc((ou & 0xFF0000)>> 8*2,f); where++;
				putc((ou & 0xFF00)>> 8,f); where++;
				putc((ou & 0xFF),f); where++;
			}
		}
		putc(nbrchar,f); where++;
		putc(POST_POST,f); where++;
		putc(postptr >> 8*3,f); where++;
		putc((postptr & 0xFF0000)>> 8*2,f); where++;
		putc((postptr & 0xFF00)>> 8,f); where++;
		putc((postptr & 0xFF),f); where++;
		putc(BZR_ID,f); where++;
		fclose(f);
	}
	return tclcode;
}


/**
 * Sbzr_LoadFont --
 * 	loads a font and sends back a list representation for it.
 *  See the header of this file for a complete description of
 *  the Tcl list returned.
 */

PRIVATE int
Sbzr_LoadFont(ClientData clientData,
							Tcl_Interp *interp,
							int objc,
							struct Tcl_Obj * CONST objv[])
{
	Tcl_Obj * res;
	if (objc < 2)
		{
			Tcl_AppendResult(interp, Tcl_GetString(objv[0]), " FontFilename", (char *) NULL);
			return TCL_ERROR;
		}
	
	
	
	if (read_font(interp, Tcl_GetString(objv[1]), &res) == TCL_OK)
		{
			Tcl_SetObjResult(interp, res);
			return TCL_OK;
		}
	else
		{
			Tcl_AppendResult(interp, Tcl_GetString(objv[1]), " does not contain a proper bzr font", (char *) NULL);
			return TCL_ERROR;
		}
}


PRIVATE int
Sbzr_SaveFont(ClientData clientData,
							Tcl_Interp *interp,
							int objc,
							Tcl_Obj * CONST *objv)
{
	if (objc != 3)
		{
			Tcl_AppendResult(interp, Tcl_GetString(objv[0]), " FONT FILENAME", (char *) NULL);
			return TCL_ERROR;
		}
	write_fonte(interp, objv[1], Tcl_GetString(objv[2]));
	return TCL_OK;
}


PRIVATE int
Sbzr_PathWinding(ClientData clientData,
								 Tcl_Interp *interp,
								 int objc,
								 Tcl_Obj * CONST *objv)
{
	int result;
	int tclcode;
	if (objc != 2)
		{
			Tcl_AppendResult(interp, Tcl_GetString(objv[0]), " PATH\n", (char *) NULL);
			Tcl_AppendResult(interp, "returns a number indicating if a path is positive or negative",
											 (char *) NULL);
			return TCL_ERROR;
		}
	tclcode= computeWinding(interp, objv[1], &result);
	if (tclcode == TCL_OK)
		{
			Tcl_Obj *res;
			res= Tcl_NewIntObj(result);
			Tcl_SetObjResult(interp, res);
		}
	return tclcode;
}

EXPORT(int,Sbzr_Init)(interp)
     Tcl_Interp *interp;
{
	/*
	 * sends back a Tcl list representing the font.
	 */
	
  Tcl_CreateObjCommand(interp, 
											 "sbzr_loadfont",
											 Sbzr_LoadFont,
											 (ClientData)NULL, 
											 (Tcl_CmdDeleteProc*) NULL);

	/*
	 * takes as argument a tcl list of the same form as the one created by
	 * sbzr_loadfont, and writes it as a bzr font.
	 */
	
  Tcl_CreateObjCommand(interp, "sbzr_savefont",Sbzr_SaveFont,
											 (ClientData)NULL, (Tcl_CmdDeleteProc*) NULL);

	
	/**
	 * For TCL programmer convenience, it might be usefull to be able to compute a path winding,
	 * for instance if one wants to convert a BZR font to postscript.
	 * Use : 
	 * 	set w [sbzr_pathwinding PATH]
	 *  w > 0 if path is positive, else < 0 
	 */

	Tcl_CreateObjCommand(interp, "sbzr_pathwinding", Sbzr_PathWinding,
											 (ClientData)NULL, (Tcl_CmdDeleteProc*) NULL);

	/*
	 * This is a new Sbzr package. Increase the package id.
	 */
	return Tcl_PkgProvide(interp, "Sbzr", "2.1");
  return TCL_OK;
}


/*
 * Local Variables:
 * tab-width: 2
 * End:
 */

