
/**
 * Code for hieroglyphic text display
 */

/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: affsigne.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: affsigne.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 * Revision 1.29  2000/12/27 12:00:51  rosmord
 * Log and id inserted
 *
 */

/*
  IDEAS
	
  * a cadrat's dimensions are first computed, 
  (via compute_texte_dimensions)
  and then the cadra can be displayed, its width written, etc.
  
  * The Texte component will ensure compute_cadrat_dimensions is
	called before anything else is done.
	
	* The width and height values kept in the structures do not
	take scale into account

	* "Natural width" or height is the width a quadra would have if it stood alone.

	* TWO sets of dimensions should be precomputed : line and column

	* Any dimension or value which can be precomputed should be so.

	* maxh belongs to the context, and should not be an argument for any function here.
*/

#include "tgraph.h"
#include "bfont.h"
#include "ligatur.h"
#include "quadra.h"
#include "seshaux.h"
#include "tksinter.h"
#include "tks_geometry.h"
#include <assert.h>
#include "affsigne.h"

/* Normally, any Non-static function from this package should set
	 these two values */

static TKSESH_output* output;
static TKSTexteGeometry *context;

#define CARTOUCHESCALE 0.8

/*
	Various Notes :
	
	a) basic idea. At the time being, a cadra is uniformaly scaled
		(sub-cadrats have their own scale, though). A cadrat made of only
		one level of signs has no width restriction (when I support
		column, this will be a problem -- I will need to "exchange" width
		and height restrictions).

		It might be interesting, in fact, to add a level-by-level scale
		(as done in HieroTeX). Both approaches have problems and merits.

	 b) from version 1.25 onward, I will try to improve parenthesis and
      cartouches handling and make it more compatible.

			Well, problems are mainly for [[ ... ]]
			Other signs are both rarer and simpler. For instance, {} or <> 
			are mainly used for either one sign or a whole cadrat.

			The problem is :
			* In winglyph, [[ ... ]] are used visually (for example, one may find
			a [[ which correspond to two ]].

			* In winglyph, [[ and ]] are considered as SIGNS.

			* The best way to go would be to consider them as toggle. A zone
			between [[ and ]] would be seen as "missing". The frontier of
			the zone would be "decorated" with [ and ].

			let's look at [[-mn:n:nw*]]*nw*nw:Z2- for example.

			[ MMMMM ]
			[ /\/\/ ]
			[ O ] O  O 
        I I I

				The missing signs are : mn, n, and nw.

				We have two different cases : external and internal.
				External : let there be a list of 
				         1) all first signs on a line
								 2) all last signs on a line

        if some of the said signs are marked as "missing", then we
        need to draw a "[" or a "]".  We then reserve some horizontal
        space for it. The drawing is then very simple : draw lines for
        each vertical connex compound, with little hooks at the ends.
				
				If the whole cadrat is affected, we must also look at the next
				cadrat before deciding : if the whole height of the cadrat is
				"missing", and the next (resp. previous) cadrat start-side
				(resp. end-side) likewise, nothing is to be
				draw. (e.g. [[-m-p-]]). 

				Then, for each sign : if the sign is "missing", and the
				previous (resp. the next) sign on the line is not, we need to
				draw a begin (resp. end) of lacuna. It's the case for the "nw"
				sign in the example. This is very simple.

				This system would give both very good visual results, be
				logical, and allow a simple and precise marking of missing
				parts.
				
				-[[-(mn:n)*]]:nw*]]*nw*nw:n-

*/

/**********************************************************************
 * Various little utilitary functions */



/* should really go somewhere else */
static double 
min(double a, double b)
{
	if (a < b)
		return a;
	else
		return b;
}

static double 
max(double a, double b)
{
	if (a > b)
		return a;
	else
		return b;
}

/**********************************************************************
 * testing functions
 */

static int
is_a_long_sign(Signe * s);

static int
is_scalable(Quadra *q);

static int
quadra_is_composite(Quadra *q);

static double 
wd_signe(Signe *s); 				/* Natural width of a sign */

static double 
ht_signe(Signe * s);

static double
wd_liggroup(struct Liggroup *l);

static double
ht_liggroup(struct Liggroup *l);

static void
compute_vertlist_dimensions(Vertlist *v);

/*
 * maxWidth : maximal width. If 0 => infinity
 */
static void
compute_cadrat_dimensions(Quadra *q, double maxWidth);

static void
compute_texte_dimensions(Texte *t);

/* 
 * is_a_long_sign --
 * does a sign belong to the category of "low, narrow" signs
 * (in which case it can be stretched to fit a cadrat's width)  
*/


static int
is_a_long_sign(Signe * s)
{
	double w;
	if (S_TYPE_OF(*s) != S_hieroglyphe)
		return FALSE;								/* should NEVER happen.  */
	w= wd_signe(s);
	if (S_LONG_OF(*s) == 1)
		return TRUE;
	if (( w > 14.0) && (w > 2.0*ht_signe(s)))
		return TRUE;
	else
		return FALSE;
}

/* is_scalable --
 * some quadra can be scaled (because they contain hieroglyphs,
 * some don't, those which contain text -- until we introduce bzr fonts
 * for text writing, that is. --
 */

static int
is_scalable(Quadra *q)
{
	if (!q) return FALSE;
	if (q->type == S_signe)
		{
			Typesigne t;
			t= S_TYPE_OF(*QD_SIGNE_OF(*q));
			if (t == S_romain || t == S_textesuper)
				return FALSE;
			else 
				return TRUE;
		}
	else
		return TRUE;
}

/* quadra_is_composite --
 *	 does an external quadra have a maximum width
 *	 due to its constitution 
 *	 (i.e. is it made of more than one row of signs ?
 */

static int
quadra_is_composite(Quadra *q)
{
	Vertlist *h;
	if (!q)
		return FALSE;
	if (S_TYPE_OF(*q) != S_quadra)
		return FALSE;
	h= QD_VERTLIST_OF(*q);
	if (h && S_NEXT_OF(*h) != NULL)
		return TRUE;
	else 
		return FALSE;
}

/**********************************************************************
 * Size computing functions 
 */

static double 
wd_signe(Signe *s)				/* Natural width of a sign. */
{
	switch(S_TYPE_OF(*s))
	{
	case S_hieroglyphe:
		switch (S_ROTATION_OF(*s))
			{
			case 0 :
			case 180 :
			case 360 :
				return 
					(S_SIZE_OF(*s) *
					 BZR_charWidth(S_CHAR_OF(*s)))/ 100.0;
				break;
			case 90 :
			case 270 :
				return 
					(S_SIZE_OF(*s)*
					 BZR_charHeight(S_CHAR_OF(*s)))/ 100.0;
				break;
			default:									/* AT THE TIME BEING, WE WON'T SUPPORT GENERAL ROTATION */
				return 
					(S_SIZE_OF(*s)*
					 BZR_charWidth(S_CHAR_OF(*s))) / 100.0;
			}
		break;
	case S_overwrite:
		return
			(max(wd_signe(S_FIRST_SIGN_OF(*s)),
					 wd_signe(S_SECOND_SIGN_OF(*s))) * (S_SIZE_OF(*s))) / 100.0;
		break;
	case S_hachureg:
	case S_hachureh:
	case S_point:
		return TKSGeom_getNaturalWidth(context);
		break;
	case S_hachurev:
	case S_hachuret:
	case S_pointpoint:
		return TKSGeom_getNaturalWidth(context)/2.0;
		break;
	case S_pointnoir:
	case S_pointrouge:
		return 
			(S_SIZE_OF(*s) *
			 BZR_charWidth(S_CHAR_OF(*s)))/ 100.0;
		break;
	case S_inconnu:
		return TKSGeom_getNaturalWidth(context);
		break;
	case S_romain:
	case S_textesuper :
		return TKSO_String_width(output, S_TEXTE_OF(*s));
		break;
	default:
		return TKSGeom_getNaturalWidth(context);
	}
}




static double 
ht_signe(Signe * s)
{
	switch(S_TYPE_OF(*s))
		{
		case S_hieroglyphe:
			switch (S_ROTATION_OF(*s))
				{
				case 0 :
				case 180 :
				case 360 :
					return 
						( S_SIZE_OF(*s)*
						 BZR_charHeight(S_CHAR_OF(*s)))/100.0;
					break;
				case 90 :
				case 270 :
					return 
						(S_SIZE_OF(*s)*
						 BZR_charWidth(S_CHAR_OF(*s)))/ 100.0;
					break;
				default:									/* AT THE TIME BEING, WE WON'T SUPPORT GENERAL ROTATION */
					return 
						(S_SIZE_OF(*s)*
						 BZR_charHeight(S_CHAR_OF(*s))) / 100.0 ;
				}
			break;
		case S_overwrite:
			return
				(max(ht_signe(S_FIRST_SIGN_OF(*s)),
						 ht_signe(S_SECOND_SIGN_OF(*s))) * (S_SIZE_OF(*s))) / 100.0;
		case S_hachureg:
		case S_hachurev:
		case S_point:
			return TKSGeom_getNaturalHeight(context);
			break;
		case S_hachureh:
		case S_hachuret:
		case S_pointpoint:
			return TKSGeom_getNaturalHeight(context)/2.0;
			break;
		case S_pointnoir:
		case S_pointrouge:
			return 
				(S_SIZE_OF(*s)*
				 BZR_charWidth(S_CHAR_OF(*s)))/ 100.0;
			break;
		default:
			return 1.0;
		}
}


static double
wd_liggroup(struct Liggroup *l)
{
	double res;
	/* 
		 If we know how to draw l, 
		 l->l contains the width.
	*/

	if (l->l)
		{
			res= l->l->w;
		}
	else
		{
			/*
				Else signs will be drawn side by side.
			 */
			struct _Signinlig *s;
			res= 0;
			s= l->signs;
			while (s)
				{
					res+= wd_signe(s->sign);
					s= s->next;
				}
		}
	return res;
}

static double
ht_liggroup(struct Liggroup *l)
{
	double res;
	/* 
		 If we know how to draw l, 
		 l->l contains the height
	*/
	if (l->l)
		{
			res= l->l->h;
		}
	else
		{
			/*
				Else signs will be drawn side by side.
			 */

			struct _Signinlig *s;
			res= 0;
			s= l->signs;
			while (s)
				{
					double tmp;
					tmp= ht_signe(s->sign);
					if (tmp > res)
						res= tmp;
					s= s->next;
				}
		}
	return res;
}

static void
compute_vertlist_dimensions(Vertlist *v)
{
	Vertlist *vv;
	double w, h, h_space;
	double sc1, sc2, sc;
	double maxWidthForLine;
	w= 0.0;
	h= 0.0;


	v->n= 0;
	
	/* 
	 * First, compute the number of elements in this vertical stack :
	 */
	
	vv= v;
	while(vv)
	{
		v->n++;
		vv=S_NEXT_OF(*vv);
	}
	
	/**
	 * if there is only one element, 
	 * we don't constrain its width
	 */
	if (v->n== 1)
		maxWidthForLine= 0.0;
	else
		maxWidthForLine= TKSGeom_getNaturalWidth(context);
	
	/* the dimensions of a verlist : */
	/* first, get the "natural" dimensions of all lines */
	vv= v;
	while(vv)
	{
		Quadra *q;
		q= VL_QUADRA_OF(*vv);
		/* The code below is a quick hack to test a few ideas. false of course */
		if (!vv->kerning && !(vv->next && vv->next->kerning) && v->n != 1)
			maxWidthForLine= TKSGeom_getNaturalWidth(context); 
		else 
			{
				maxWidthForLine= 0.0;
			}
		compute_cadrat_dimensions(q, maxWidthForLine);
		w= max(q->w,w);
		h+= q->h;
		vv=S_NEXT_OF(*vv);
	}

	/* Then, play with them to fit the normal dimensions */
	h_space= h;

	/* we should have some spacing between lines : */	
	if (v->n > 1)
		h_space+= TKSGeom_getSmallSkip(context) * (v->n - 1);

	if (h_space> TKSGeom_getNaturalHeight(context))
		sc2= TKSGeom_getNaturalHeight(context)/h_space;
	else
		sc2= 1.0;

	/**
	 * This code has normally no use. 
	 * we keep it just in case we use it again.
	 */
	if (w > TKSGeom_getNaturalWidth(context))
		sc1= TKSGeom_getNaturalWidth(context)/w;
	else 
		sc1= 1.0;
	/* sets sc1 to 1.0 in all cases. */
	sc1= 1.0;

	sc= min(sc1,sc2);

	/* finaly, get the vertical list height and width */
	v->h= h_space*sc;
	v->hcompact= h*sc;
	v->w= w*sc;
	v->scale= sc;
}

/**
 *
 */
static void
compute_cadrat_dimensions(Quadra *q, double maxWidth)
{
	Quadra *qq;
	double w, h;
	w=0.0;
	h=0.0;

	q->n= 0;
	qq= q;
	while (qq)
		{
			double subw, subh;
			q->n++;
			switch (S_TYPE_OF(*qq))
				{
				case S_signe:
					subw= wd_signe(QD_SIGNE_OF(*qq));
					subh= ht_signe(QD_SIGNE_OF(*qq));
					break;
				case S_quadra:
					compute_vertlist_dimensions(QD_VERTLIST_OF(*qq));
					subw= QD_VERTLIST_OF(*qq)->w;
					subh= QD_VERTLIST_OF(*qq)->h;					
					break;
				case S_carto:
				case S_debcarto:
				case S_milcarto :
				case S_fincarto :
				case S_serekh :
				case S_enceinte :
				case S_chateau :
				case S_debserekh :
				case S_milserekh :
				case S_finserekh :
				case S_debenceinte :
				case S_milenceinte :
				case S_finenceinte :
				case S_debchateau :
				case S_milchateau :
				case S_finchateau :
					{
						/* First, compute the dimensions of the cartouche content */
						compute_cadrat_dimensions(QD_QUADRA_OF(*qq), 0.0);
						subw= QD_QUADRA_OF(*qq)->w * CARTOUCHESCALE + 14.0;
						subh= QD_QUADRA_OF(*qq)->h ;
					}
					break;
				case 	S_superfetatoire:
				case S_efface:
				case S_disparu:
				case S_ajoutscribe:
				case S_ajoutauteur:
					compute_cadrat_dimensions(QD_QUADRA_OF(*qq),maxWidth);
					subw= QD_QUADRA_OF(*qq)->w + 8.0;
					subh= QD_QUADRA_OF(*qq)->h ;
					break;
				case S_ligature:
					subw= wd_liggroup(QD_LIG_OF(*qq));
					subh= ht_liggroup(QD_LIG_OF(*qq));
					break;
				case S_rightleft:
					subw= 0.0;
					subh= 0.0;
					break;
				}
			w+= subw;
			h= max(h, subh);
			qq= S_NEXT_OF(*qq);
		}

	q->h=h;	
	q->wcompact= w;
	q->w= w;
	if (q->n > 1) {
		q->w+= TKSGeom_getSmallSkip(context) * (q->n - 1);
		if (maxWidth != 0.0 && w > TKSGeom_getNaturalWidth(context))
			q->scale= TKSGeom_getNaturalWidth(context) / w;
	}

	q->w= q->w*q->scale;
	q->wcompact= q->wcompact*q->scale;
}

/**
 * The goal of this function is to compute the real text dimensions.
 * Before : t, a cadrat without its real dimensions
 * After : t's dimensions are sound.
 */

static void
compute_texte_dimensions(Texte *t)
{
	if (t)
		{
			if (t->q)
				{
					compute_cadrat_dimensions(t->q, FALSE);
					/* Now, the previous dimensions do not take 
						 the real requested height into account.
						 Let's scale the whole stuff */

					if (is_scalable(t->q))
						{
							/* First, deal with the vertical case : */
							double ht, wd, scale, scale1;

							ht= t->q->h;
							wd= t->q->w;

							/* We get sure ht and wd are large enough.
							 *	 That is, smaller cadrats will get more spaces
							 */
							if ( ht < TKSGeom_getNaturalHeight(context))
								ht= TKSGeom_getNaturalHeight(context);
							if ( wd < TKSGeom_getNaturalWidth(context))
								wd= TKSGeom_getNaturalWidth(context);
								
							/*
								The primary factor for size is horizontal dimension
							*/

							scale= TKSGeom_getMaxW(context)/wd;

							t->wv= t->q->w*scale + 0.1*TKSGeom_getMaxW(context);
							t->hv= t->q->h*scale;
							t->scalev= scale;

							/* Now, deal with dimensions in line oriented-text */
							
							
							/* scale is the scale needed to fix vertical dimensions */
							scale= TKSGeom_getMaxH(context)/ht;
							/* Now, scale1 is the scale to use to fix horizontal dimensions, and keep
							 *	 a good ratio.
							 */
							scale1= (TKSGeom_getNaturalWidth(context)/TKSGeom_getNaturalHeight(context))*TKSGeom_getMaxH(context)/wd;
							/* In case the cadrat is too wide with scale, we use scale1 */ 
							/* THIS RULE HAS BEEN SUPRESSED. IT WASN'T USEFUL ! */  
							/* if (quadra_is_composite(t->q) && scale1 < scale)
								 scale= scale1;*/
							t->wh= t->q->w*scale + 0.1*TKSGeom_getMaxW(context);
							t->hh= TKSGeom_getMaxH(context);
							t->scaleh= scale;
						} else {
							/* text item can't be easily scaled*/
							t->wv= t->wh= t->q->w + 0.1*TKSGeom_getMaxW(context);
							t->hv= t->hh= t->q->h;
							t->scalev= t->scaleh= 1.0;
						}
				}
			else
				{
					t->wv= t->wh= 0;
					t->hv= t->hh= 0;
					t->scalev= t->scaleh= 1.0; /* just in case */
				}
			t->real_dims= TRUE;
		}
}


/**********************************************************************
 * Drawing functions
 */


static void
drawShading(double x, double y,double w, double h, Hachures hach);

/**
 * Draws a more or less square shading, given a position, and a scale. 
 */

static void
drawShadingScale(double x, double y, double sx, double sy, Hachures hach);

static void
inverser(double x, double y,double w, double h);

static void
affiche_parentheses(double x0, double x1, double y, double h, double echelle, Typegroupe type);

static void
dessine_cartouche(TKSESH_rectangle cart, double echelle, Typegroupe type);

static void 
aff_signe(
					double *x,
					double *y,
					Signe *s,
					double tx, double ty);

static void
aff_liggroup(double *x, double *y,
						 struct Liggroup *l,
						 double echelle);

static void
aff_vertlist(double *x,
						 double *y, 
						 Vertlist *v,
						 double scale,
						 double htmax);

static void 
real_aff_quadra(double *x,			/* left coordinate */
								double *y,			/* upper y coordinate */
								Quadra *q,			/* cadrat to draw */
								double wd,			/* actual width */
								double htmax,		/* actual max. height */
								double scale,		/* scale to use for this purpose */
								int external);	/* outer-level ? */


static void
drawShading(double x, double y,double w, double h, Hachures hach)
{
	if (hach==NOHACHURE)
		return;
	else
		{
			TKSESH_rectangle grise;
			if ((hach)&(HACHURE1))
				{
					grise= TKSO_init_rectangle(x,y-h/2,x+w/2,y);
					TKSO_hachure(output, grise);
				}
			if ((hach)&(HACHURE2))
				{
					grise= TKSO_init_rectangle(x+w/2,y-h/2,x+w,y);
					TKSO_hachure(output, grise);
				}if ((hach)&(HACHURE3))
				{
					grise= TKSO_init_rectangle(x,y-h,x+w/2,y-h/2);
					TKSO_hachure(output, grise);
				}if ((hach)&(HACHURE4))
				{
					grise= TKSO_init_rectangle(x+w/2,y-h,x+w,y-h/2);
					TKSO_hachure(output, grise);
				}
		}
}

static void
drawShadingScale(double x, double y, double sx, double sy, Hachures hach)
{
	double dimx, dimy;
	dimx= TKSGeom_getNaturalWidth(context);
	dimx= TKSGeom_getNaturalHeight(context);
	drawShading(x, y+sy*dimy, sx*dimx, sy*dimy, HACHURE1|HACHURE2);
}


static void
inverser(double x, double y,double w, double h)
{
	TKSESH_rectangle grise;
	TKSO_state st;
	st= TKSO_get_state(output);
	TKSO_set_state(output, black);
	grise= TKSO_init_rectangle(x,y-h,x+w,y);
	TKSO_fill_rectangle(output, grise);
	TKSO_set_state(output, st);
}


static void
affiche_parentheses(double x0, double x1, double y, double h, double echelle, Typegroupe type)
{
	switch (type) {
	case 	S_superfetatoire:
		{
			/* Should be an accolade. Draw it if you will */
			TKSESH_point 
				tb, te, /* top begin and end */
				m1b, m1e, /* straight portion of the accolade*/
				m2b, m2e,
				c1b, c1e, /* center ...*/
				c2b, c2e, /* center ...*/
				m3b, m3e, /* straight portion of the accolade*/
				m4b, m4e,
				bb, be; /* bottom */
			double dimArc= echelle*2;
			/* start of the accolade : an arc. */ 
			tb.x= (x0+ dimArc);
			tb.y= (y + h);
			m1b.x= x0+ dimArc;
			m1b.y= y + h - dimArc;
			m2b.x= m1b.x;
			m2b.y= y + h/2.0 + dimArc;
			TKSO_DrawArc(output, tb, dimArc, dimArc, 90*64, 90*64);
			TKSO_DrawLine(output, m1b, m2b);
			/* middle point */
			c1b.x= x0;
			c1b.y= y + h/2.0 + dimArc;
			TKSO_DrawArc(output, c1b, dimArc, dimArc, 270*64, 90*64);
			m3b.x= m1b.x;
			m4b.x= m1b.x;
			m3b.y= y+ h/2 - dimArc;
			m4b.y= y+ dimArc;
			c2b.x= x0;
			c2b.y= y+h/2;
			TKSO_DrawArc(output, c2b, dimArc, dimArc, 0*64, 90*64);
			TKSO_DrawLine(output, m3b, m4b);
			bb.x= x0 + dimArc;
			bb.y= y+dimArc; 
			TKSO_DrawArc(output, bb, dimArc, dimArc, 180*64, 90*64);
			/* Second accolade */
			te.x= (x1);
			te.y= (y + h);
			m1e.x= x1+ dimArc;
			m1e.y= y + h - dimArc;
			m2e.x= m1e.x;
			m2e.y= y + h/2.0 + dimArc;
			TKSO_DrawArc(output, te, dimArc, dimArc, 0*64, 90*64);
			TKSO_DrawLine(output, m1e, m2e);
			c1e.x= m2e.x;
			c1e.y= m2e.y;
			TKSO_DrawArc(output, c1e, dimArc, dimArc, 180*64, 90*64);
			c2e.x= c1e.x;
			c2e.y= y+h/2.0;
			TKSO_DrawArc(output, c2e, dimArc, dimArc, 90*64, 90*64);
			m3e.x= m1e.x;
			m4e.x= m1e.x;
			m3e.y= y+ h/2 - dimArc;
			m4e.y= y+ dimArc;
			TKSO_DrawLine(output, m3e, m4e);
			be.x= x1;
			be.y= y+ dimArc;
			TKSO_DrawArc(output, be, dimArc, dimArc, 270*64, 90*64);
		}
		break;	case S_ajoutscribe:
		{
			TKSESH_point 
				ah, ab, bh, bb;
			ah.x= x0 + 2.0*echelle;
			ah.y= y+h;
			ab.x= ah.x;
			ab.y= y+h*3/4.0;
			bh.x= x1 + 2.0 * echelle;
			bh.y= ah.y;
			bb.x= bh.x;
			bb.y= ab.y;
			TKSO_DrawLine(output, ah, ab);
			TKSO_DrawLine(output, bh, bb);
		}
		break;
	case S_ajoutauteur:
		{
			TKSESH_point 
				tb, te, /* top begin and end */
				cb, ce, /* center ...*/
				bb, be; /* bottom */
			tb.x= (x0 + 2.0 * echelle);
			tb.y= (y);
			cb.x= x0;
			cb.y= y+h/2;
			bb.x= tb.x;
			bb.y= (y+h);
			TKSO_DrawLine(output, tb, cb);
			TKSO_DrawLine(output, cb, bb);

			te.x= (x1);
			te.y= (y);
			ce.x= x1 + 2.0 * echelle;
			ce.y= y+h/2;
			be.x= te.x;
			be.y= (y+h);
			TKSO_DrawLine(output, te, ce);
			TKSO_DrawLine(output, ce, be);
		}
		break;
	case S_disparu: /* drawn like that : [| ... |]. */
		{
			TKSESH_point 
				ah, ab, bh, bb;
				ah.x= x0 + 2.0 * echelle;
				ah.y= y+h/4.0;
				ab.x= ah.x;
				ab.y= y+3*h/4.0;
				bh.x= x1 + 2.0 * echelle;
				bh.y= ah.y;
				bb.x= bh.x;
				bb.y= ab.y;
				TKSO_DrawLine(output, ah, ab);
				TKSO_DrawLine(output, bh, bb);
		}
	/* no break : S_efface will draw the square brackets*/
	case S_efface:
		{
			TKSESH_point 
				aa, ab, ac, ad,
				ba, bb, bc, bd;
			aa.x= (x0);
			aa.y= (y);
			ab.x= (x0 + 2.0 * echelle);
			ab.y= (y);
			ac.x= aa.x;
			ac.y= (y+h);
			ad.x= ab.x;
			ad.y= ac.y;
			TKSO_DrawLine(output, aa, ab);
			TKSO_DrawLine(output, aa, ac);
			TKSO_DrawLine(output, ac, ad);

			bb.x= (x1 + 2.0 * echelle);
			bb.y= (y);
			ba.x= (x1 + 4.0 * echelle);
			ba.y= (y);
			bc.x= ba.x;
			bc.y= (y+h);
			bd.x= bb.x;
			bd.y= bc.y;
			TKSO_DrawLine(output, ba, bb);
			TKSO_DrawLine(output, ba, bc);
			TKSO_DrawLine(output, bc, bd);
		}
		break;
	default:
		return;
	}
}


static void
dessine_cartouche(TKSESH_rectangle cart, double echelle, Typegroupe type)
{
	double echellef, width, height, half_width;

	echellef= (echelle);
	
	width= (echelle*7.0);
	half_width= (echelle*7.0/2.0);
	height= cart.top_right.y - cart.bottom_left.y;

	if (type == S_carto ||
			type == S_debcarto||
			type == S_rightleft||
			type == S_milcarto ||
			type == S_fincarto)
		{
			TKSESH_point 
				ac, bc,
				ah, ab,
				bh, bm, bb,
				th, tb;
			
			ah.x= cart.bottom_left.x + half_width;
			ah.y= cart.top_right.y;
			ab.x= ah.x;
			ab.y= cart.bottom_left.y;
			bh.x= cart.top_right.x - half_width;
			bh.y= ah.y;
			bb.x= bh.x;
			bb.y= ab.y;
			bm.x= cart.top_right.x;
			bm.y= (ah.y + ab.y)/ 2.0;
			th.x= tb.x= bm.x;
			th.y= bh.y;
			tb.y= bb.y;

			/* upper left corner for the left arc */
			ac.x= cart.bottom_left.x;
			ac.y= cart.top_right.y;
			/* upper left corner for the right arc */
			bc.x= bb.x - half_width;
			bc.y= ac.y;
		
			if (type != S_milcarto && type != S_fincarto)
				{
					/* the cartouche's beginning */
					TKSO_DrawArc(output, ac, width, height, 90*64, 180*64);
				}
			if (type != S_milcarto && type != S_debcarto)
				{
					/* the cartouche's ending */
					TKSO_DrawArc(output, bc, width, height, 270*64, 180*64);
					/* the cartouche's node */
					TKSO_DrawLine(output, th, tb);
				}
			TKSO_DrawLine(output, ah, bh);
			TKSO_DrawLine(output, ab, bb);
		}
	if (type == S_chateau ||
			type == S_debchateau ||			
			type == S_milchateau ||
			type == S_finchateau) 
		{
			TKSESH_point 
				a, b, a1, b1, c1;
			a.x= cart.bottom_left.x;
			a.y= cart.top_right.y;
			b.x= cart.top_right.x;
			b.y= cart.bottom_left.y;

			/* frame */
			TKSO_DrawLine(output, a, cart.bottom_left);
			TKSO_DrawLine(output, cart.bottom_left, b);
			TKSO_DrawLine(output, b, cart.top_right);
			TKSO_DrawLine(output, cart.top_right, a);

			/* subframe */
			a1.x= cart.bottom_left.x + half_width;
			a1.y= cart.bottom_left.y;
			b1.x= cart.bottom_left.x;
			b1.y= cart.bottom_left.y + half_width;
			c1.x= a1.x;
			c1.y= b1.y;
			TKSO_DrawLine(output, a1, c1);
			TKSO_DrawLine(output, c1, b1);
		}
	
	if (type == S_serekh
			|| type == S_debserekh
			|| type == S_milserekh
			|| type == S_finserekh)
		{
			/* a         b
			 * a1        b1 */
			int i;
			TKSESH_point 
				a, b, a1, b1, c1, c2;
			double step;
			
			a1= cart.bottom_left;
			a.x= cart.bottom_left.x;
			a.y= cart.top_right.y;
			b.x= cart.top_right.x - half_width;
			b.y= cart.top_right.y;
			b1.x= b.x;
			b1.y= cart.bottom_left.y;
			
			/* frame */
			TKSO_DrawLine(output, a, b);
			TKSO_DrawLine(output, b, b1);
			TKSO_DrawLine(output, b1, a1);
			TKSO_DrawLine(output, a1, a);

			/* palace-facade lines */
			c1.x= b.x;
			c1.y= b.y;
			c2.x= cart.top_right.x;
			c2.y= c1.y;
			step= (b.y - b1.y) / 5.0;
			for (i= 0; i < 5; i++)
				{
					TKSO_DrawLine(output, c1, c2);
					c1.y= c1.y - step;
					c2.y= c1.y;
				}
			c1.y= b1.y;
			c2.y= c1.y;
			TKSO_DrawLine(output, c1, c2);
		}
}



/* before calling aff_signe, x is the abciss of the sign, and y is its top */
/* After aff_signe, *x will be after the sign, and *y will be at the sign's basis */
/* t is the sign's scale */
static void 
aff_signe(
					double *x,
					double *y,
					Signe *s,
					double scalex, double scaley)
{
	double tx, ty; /* Final scale, using both external scale (scalex and scaley),
									*	and intrinsic sign scale (size).
									*/
	TKSESH_point m;
	TKSO_state st;

	if (S_EST_SELECTED(*s)) {
		inverser(*x,
						 *y /* +ht_signe(s)*ty */,
						 wd_signe(s)*scalex,
						 ht_signe(s)*scaley);
		TKSO_toggle_reverse(output);		
	}

	/* set scaling factor */
	/* put *y at the at the base of sign t */
	*y-= ht_signe(s)*scaley;

	tx= (scalex * S_SIZE_OF(*s))/100.0;		
	ty= (scaley * S_SIZE_OF(*s))/100.0;		

	m= TKSO_init_point(*x, *y);
	/* possible color change */

	st= TKSO_get_state(output);
	if (TKSGeom_getShowWords(context) && S_EST_FIN_MOT(*s))
		{
			TKSO_set_state(output, blue);
		}
	if (TKSGeom_getShowWords(context) && S_EST_GRAMM(*s))
		{
			TKSO_set_state(output, gram);
			if (TKSGeom_getShowWords(context) && S_EST_FIN_MOT(*s))
							TKSO_set_state(output, gramend);
		}
	if (S_IGNORE_OF(*s))
		TKSO_set_state(output, ignore);

	/* real sign writting  */

	switch(S_TYPE_OF(*s))
		{
		case S_hieroglyphe:
			BZR_putc(S_CHAR_OF(*s), 
							 tx, ty,
							 S_ROTATION_OF(*s),
							 S_REVERT_OF(*s) && ! TKSGeom_isRightLeftP(context),
							 output, m);
			break;
		case S_overwrite:
			{
				double x1, x2, y1, y2;
				double ht1, ht2, htmax;
				double w1, w2;
						
				/* y goes back to where it was. We should make this 
					 cleaner */
				*y+= ht_signe(s)*scaley;
						
				x2= x1= *x;
				y2= y1= *y;
				ht1= ht_signe(S_FIRST_SIGN_OF(*s))* scaley;
				ht2= ht_signe(S_SECOND_SIGN_OF(*s))* scaley;
				w1= wd_signe(S_FIRST_SIGN_OF(*s)) * scalex;
				w2= wd_signe(S_SECOND_SIGN_OF(*s))* scalex;
				htmax= max(ht1,ht2);
					
				if (ht2> ht1* 1.5)	 /* 1st sign centered */
					{
						y1-= (htmax- ht1)/2.0;
						y2-= (htmax- ht2);
					} 
				else if (ht1 > ht2*1.5)	/* 2nd sign centered */
					{
						y1-= (htmax- ht1);
						y2-= (htmax- ht2)/2.0;
					}
				else								/* signs on the same line */
					{
						y1-= (htmax- ht1);
						y2-= (htmax- ht2);
					}
						
				if (w1 > w2*1.5) 
					{
						x2+= (w1-w2)/2.0;
					}
				else if (w2 > w1*1.5) 
					{
						x1+= (w2-w1)/2.0;
					}

				aff_signe(&x1,&y1,S_FIRST_SIGN_OF(*s),tx,ty);
				aff_signe(&x2,&y2,S_SECOND_SIGN_OF(*s),tx,ty);
				if (y1 > y2)
					*y= y2;
				else
					*y= y1;
			}
					break;
		case S_inconnu:
		case S_romain :
			{
				if (! TKSGeom_isInColumnsP(context)) {
					TKSESH_point mcurr;
					mcurr= TKSO_init_point(*x, *y);				
					TKSO_DrawString(output, mcurr, S_TEXTE_OF(*s));
				}
			}
			break;
		case S_textesuper :
			{
				if (! TKSGeom_isInColumnsP(context)) {
					TKSESH_point mcurr;
					mcurr= TKSO_init_point(*x , *y + 0.5);				
					TKSO_DrawString(output, mcurr, S_TEXTE_OF(*s));
				}
			}
		break;		
		case S_pointnoir:
			BZR_putc(S_CHAR_OF(*s),
							 tx, ty,
							 S_ROTATION_OF(*s),
							 S_REVERT_OF(*s),
							 output, m);
			break;
		case S_pointrouge:
			{
				TKSO_state st;
				st= TKSO_get_state(output);
				TKSO_set_state(output, red);
				BZR_putc(S_CHAR_OF(*s),
								 tx, ty,
								 S_ROTATION_OF(*s),
								 S_REVERT_OF(*s),
								 output, m);
				TKSO_set_state(output, st);
			break;
			}
		case S_hachureh:
			drawShadingScale(*x,*y, tx, ty, HACHURE1|HACHURE2);
			break;
		case S_hachurev:
			drawShadingScale(*x,*y, tx, ty, HACHURE1|HACHURE3);
			break;
		case S_hachuret:
			drawShadingScale(*x,*y, tx, ty, HACHURE1);
			break;
		case S_hachureg:
			drawShadingScale(*x,*y, tx, ty, TOUTHACHURE);
			break;
		default:
			break;			
		}

	/* We test if we need to restore the previous state.
		 The idea is that state changing might be expensive on some systems
	*/
	if ((TKSGeom_getShowWords(context) && S_EST_FIN_MOT(*s)) 
			|| (TKSGeom_getShowWords(context) && S_EST_GRAMM(*s))
			|| S_IGNORE_OF(*s))
		{
			TKSO_set_state(output, st);
		}
	*x+= wd_signe(s)*scalex;
	
	if (S_EST_SELECTED(*s)) 
		TKSO_toggle_reverse(output);

}




/* As noted below : at calling time,
*  *x and *y are the UPPER LEFT COORDS
*  of the text. after call they are the
*  LOWER RIGHT ones 
*/

static void
aff_liggroup(double *x, double *y,
						 struct Liggroup *l,
						 double echelle)
{
	if (l->l)
		{
			struct _Signinlig *s;
			struct sign_ligpos *sp;
			double miny;
			
			s= l->signs;
			sp= l->l->signs;
			miny= *y - l->l->h*echelle;
			while (s)
				{
					double x1= *x;
					double y1= miny;
					/* 
						 in sp, we have got :
						 x, y and size
					*/
					x1+= sp->x * echelle;
					/* for y the formula 
						 is a bit more complex :
						 miny + sp->y + SIGN HEIGHT
					*/
					y1+= (sp->y + ht_signe(s->sign)*sp->size/100.0) * echelle ;
					/* the real sign is in s */
					aff_signe(&x1, &y1,
										s->sign,
										echelle*sp->size/100.0,
										echelle*sp->size/100.0);
					s= s->next;
					sp++;
				}
			*y= miny;
			*x= *x + l->l->w* echelle;
		}
	else
		{
			struct _Signinlig *s;
			double miny;
			miny= *y;
			s= l->signs;
			while (s)
				{
					double y1= *y;
					aff_signe(x, &y1,
										s->sign,
										echelle,
										echelle);
					if (y1 < miny)
						miny=y1;
					s= s->next;
				}
			*y= miny;
		}
}						 

static void
aff_vertlist(double *x,
						 double *y, 
						 Vertlist *v,
						 double scale,
						 double htmax)
{
	int i;
	double xx;
	double ht;
	double subscale;							/* scale for embedded constructs */
	double above, sep;	
	Vertlist *vv;

	
	subscale= scale * v->scale;
	
	ht= v->hcompact*scale;
	xx= *x;
	
	if (TKSGeom_isInColumnsP(context))
		{
			above= 0.0;
			sep= TKSGeom_getSmallSkip(context)*subscale;
		}
	else {
		if (v->n == 1)
			{
				if (TKSGeom_isCenteredP(context))
					{
						above= (htmax - ht)/2.0;
						sep= above;
					}
				else
					{
						above= htmax - ht;
						sep= 0.0;
					}
			}
		else
			{
				above= 0.0;
				sep= (htmax-ht) / (v->n-1);
			}
	}
	*y-= above;

	vv= v;
	for(i=0; i<v->n; i++)
	{
		double xxx= *x;
		double haut;
		haut= (VL_QUADRA_OF(*vv))->h * subscale;
		real_aff_quadra(&xxx, y, VL_QUADRA_OF(*vv), 
										v->w*scale,
										haut,
										subscale,
										FALSE);
		if (i != (v->n-1)) *y -= sep;
		vv= S_NEXT_OF(*vv);
		if (xxx > xx) 
			xx= xxx;
	}
	*x= xx;
}

/*
 *  quadra here means an horizontal list of groups.
 *  ??? that's a rather strange idea
 */
	 
static void 
real_aff_quadra(double *x,
								double *y,
								Quadra *q,
								double wd,
								double htmax,
								double scale,
								int external)
{
  int i;
  double 
		sep,												/* white space before and after quadra */
		around;											/* white space between quadra elements */
  double yy;			/* maxy */
	Quadra *qq;
	int oneExternalSign= FALSE;
	double internalScale;

  
	/* WE DO NOT USE YET q->scale. 
		 The following line should be uncommented 
		 if we decided to use it (for example to give a different
		 scale to different lines in a cadrat)
	*/
	
	internalScale= q->scale;
	
		 
	/* Now, wd is the requested width. 
		 The "normal" width is q->w, 
		 and we should fill the rest with spaces.
	*/
	
	/* First, we compute spaces. */
	
	if (q->n == 1)
		{
      sep= 0.0;
      around = (wd - q->wcompact*scale) / 2;
    }
	else
		{
			around=0.0;
      sep= (wd-q->wcompact*scale)/(q->n-1);
    }
	
  *x+= around;

	if (q->n == 1 && external)
		oneExternalSign= TRUE;

	qq= q;
  for (i= 0, yy= *y; i<q->n; i++)
    {
			/* yy is the final position for the whole block of cadrats */
			double origx= *x;					/* starting x for the current cadrat*/
			double origy= *y;					/* starting y for the current cadrat */
      double yyy= origy;				/* at first, yyy is origy. Then it contains the final y for this cadrat */
      switch(S_TYPE_OF(*qq))
				{
				case S_signe:
					{
						/*
						 * A sign should normaly be drawn at the quadra's scale.
						 * Yet, if it's alone on his line, and it is a long, narrow,
						 * sign, it looks better if it fills the line, even if we distord it
						 * The distorsion might be huge in the case of hieratic texts, but
						 * there's no way of getting it right anyway, 
						 * plus it's quite faithful to the
						 * peculiarities of hieratic writting.
						 */
						double scalea= scale*internalScale;
						
						if (q->n == 1 && is_a_long_sign(QD_SIGNE_OF(*qq)))
							{
								*x-= around;
								around= 0.0;
								scalea= wd/wd_signe(QD_SIGNE_OF(*qq))*internalScale;
								/* in fact : scale*(wd/(width(qq)*scale)) */
							}
						yyy-= (htmax- ht_signe(QD_SIGNE_OF(*qq))* scale*internalScale);
						aff_signe(x,&yyy,QD_SIGNE_OF(*qq),scalea,scale*internalScale);
					}
					break;
				case  S_quadra:
					{
						/* embedded quadra. */
						aff_vertlist(x,&yyy, QD_VERTLIST_OF(*qq),
												 scale*internalScale,
												 htmax);
					}
					break;
				case S_carto:
				case S_debcarto:
				case S_milcarto :
				case S_fincarto :
				case S_serekh :
				case S_enceinte :
				case S_chateau :
				case S_debserekh :
				case S_milserekh :
				case S_finserekh :
				case S_debenceinte :
				case S_milenceinte :
				case S_finenceinte :
				case S_debchateau :
				case S_milchateau :
				case S_finchateau :
					{
							TKSESH_rectangle cartouche;
							double tmpx, tmpy;
							double real_h, real_w;
							double textHeight;
							int center;

							center= TKSGeom_isCenteredP(context);
							/* Hieroglyphic text should be centered in a cartouche */
							TKSGeom_setCentered(context,TRUE);

							/* real_h= q->h*scale; */
							real_w= QD_QUADRA_OF(*qq)->w*scale*CARTOUCHESCALE;
							
							textHeight= htmax*CARTOUCHESCALE;
							
							tmpx= *x+ 7.0*scale;
							/* find the vertical position of the text : */
							tmpy= yyy - ((htmax-textHeight)/2.0);
							printf("%f %f\n ", (htmax-textHeight)/2.0, tmpy);
							
							
							real_aff_quadra(&tmpx, &tmpy, QD_QUADRA_OF(*qq),
															real_w, textHeight, CARTOUCHESCALE*scale, FALSE);

							cartouche= TKSO_init_rectangle(*x, yyy - htmax, 
																								tmpx + 7.0*scale, yyy);
							dessine_cartouche(cartouche, scale, S_TYPE_OF(*qq));
							yyy= tmpy;
							*x= tmpx + 7.0*scale;
							TKSGeom_setCentered(context, center);
						}
					break;
				case 	S_superfetatoire:
				case S_efface:
				case S_disparu:
				case S_ajoutscribe:
				case S_ajoutauteur:
					{
						double tmpx, tmpy;
						double real_h, real_w;
						double finalScale= scale*internalScale;
						tmpx= *x+ 4.0* finalScale;
						tmpy= yyy;
						
						real_h= q->h*finalScale;
						real_w= QD_QUADRA_OF(*qq)->w*finalScale;
						real_aff_quadra(&tmpx, &tmpy, QD_QUADRA_OF(*qq), 
														real_w, htmax, finalScale, FALSE);
						affiche_parentheses(*x, tmpx, tmpy, real_h, finalScale, S_TYPE_OF(*qq));
						yyy= tmpy;
						*x= tmpx + 4.0* finalScale;
					}					
					break;
				case S_ligature:
					{
						yyy-= (htmax- ht_liggroup(QD_LIG_OF(*qq))* scale*internalScale);
						aff_liggroup(x,&yyy, QD_LIG_OF(*qq), scale*internalScale);
						break;
					}
				case S_rightleft:
					break;
				}
			// Now :
			//   origx and origy contain the original values for x and y
			//   *x and yyy contain the final values
			if (! oneExternalSign)
				{
					if (qq->est_grise)
						drawShading(origx,origy,*x-origx,*y - yyy,TOUTHACHURE);
					
					drawShading(origx,origy,*x-origx,*y - yyy,qq->hachures);
				}
      qq= S_NEXT_OF(*qq);
			if (i != q->n - 1) *x+= sep;	/* pas pour le dernier. */
      if (yyy < yy) yy= yyy;
    }
  *x+= around;
  *y= yy;
}

/**********************************************************************
 * PUBLIC, exported functions
 */

/*
 * SESH_fix_dims --
 * computes the dimensions of a Text item (its width and its height),
 */

void
SESH_fix_dims(TKSESH_output *where, TKSTexteGeometry *geom, Texte *t)
{
	if (!t)
		return;
	output= where;
	context= geom;
  if (!t->real_dims)
    compute_texte_dimensions(t);
	t->real_dims= TRUE;
}

double
SESH_texte_width(TKSESH_output *where, TKSTexteGeometry *geom, Texte *t)
{
  assert(t->real_dims);
	output= where;
	context= geom;
	if (TKSGeom_isInColumnsP(geom))
		return t->wv;
	else
		return t->wh;
}

double
SESH_texte_height(TKSESH_output *where, TKSTexteGeometry *geom, Texte *t)
{
  assert(t->real_dims);
	output= where;
	context= geom;
	if (TKSGeom_isInColumnsP(geom))
		return t->hv;
	else
		return t->hh;
}

static double
SESH_texte_getScale(Texte *t)
{
  assert(t->real_dims);
	if (TKSGeom_isInColumnsP(context))
		return t->scalev;
	else
		return t->scaleh;
}

void 
SESH_aff_texte(TKSESH_output *where,
							 TKSTexteGeometry *geom,
							 double *x, double *y,
							 Texte *t)
{
  /*double wd, scale;*/
  double xorig, yorig;
	double height;
  TKSO_state st;	
  /*double espace_occupe= 1.0;*/
  Quadra *q;

  if (t)
    q= t->q;
  else
    return;
	
	output= where;
	context= geom;

  if (!t->real_dims)
    compute_texte_dimensions(t);

	height= SESH_texte_height(where, geom, t);

  if (!q) return;
	
  *y+= height; /* Texte's y is lower coordinate,
								but we start drawing at the TOP of the cadrat*/
  xorig= *x;
  yorig= *y;

  /* If current quadra is selected, sets the writing style to selected
     and fills the background in black
  */

  st= TKSO_get_state(output);

  if (t->selected)
    {
      inverser(xorig,yorig,SESH_texte_width(where, geom, t),height);
      TKSO_toggle_reverse(output);
    }

  real_aff_quadra(x, y, q, SESH_texte_width(where, geom, t), height, SESH_texte_getScale(t), TRUE);
	
	drawShading(xorig,yorig,*x-xorig,height,q->hachures);
  if (q->est_grise)
		drawShading(xorig,yorig,*x-xorig,height,TOUTHACHURE);
	
  if (t->selected)
    TKSO_toggle_reverse(output);
}



/*
 * Local Variables:
 * header: "affSigne.h"
 * tab-width: 2
 * End:
 */

