#
# Glyph Data Access Layer
#
# For the part of the system which is not OO,
# this package should contain *ALL* SQL request 
# which access something related to glyphs.
# 
#
# Important notice :
#
# In this package, we use the words :
# 	glyph : for a particular drawing, saved in the base
#	sign : for an "abstract" sign, usually represented by one or more glyphs.
#	

package provide glyphDAL 1.0
package require TkseshDB

namespace eval glyphDAL {

    # Sends back a list of signs matching a given glob pattern.
    proc getSignsCodesMatching {pattern} {
	set db [TkseshDB::getTkseshDB]
	set sql "select distinct code from HasCode where code GLOB '$pattern'"
	set l [TkseshDB::select $db $sql]
	return $l
    }
    
    proc getValuesForCode {code} {
	set db [TkseshDB::getTkseshDB]
	set sql "select distinct value from HasValue where code = '$code'"
	set val [TkseshDB::select $db $sql]
	return $val
    }
    
    proc getCodesForValue {value} {
	set db [TkseshDB::getTkseshDB]
	set sql "select distinct code from HasValue where value = '$value'"
	set result [TkseshDB::select $db $sql]
    }

    proc getPhoneticCodeForSignCode {signCode} {
	set db [TkseshDB::getTkseshDB]
	set sql "select code from PhonCode where sign = '$signCode'"
	set r [TkseshDB::select $db $sql]
	return $r
    }

    # returns if 'code' is actually a phonetic code for some sign
    # (e.g. sHtp isn't an official phonetic code, whereas 'mn' is).
    proc isOfficialPhoneticCode {code} {
	set s [getSignCodeForPhoneticCode $code]
	puts "got code $s"
	return [expr [string length $s] != 0]
    }

    # returns the gardiner-like sign code for a given phonetic code,
    # or the empty string.
    proc getSignCodeForPhoneticCode {PHONCODE} {
	set db [TkseshDB::getTkseshDB]
	set sql "select sign from PhonCode where code = '$PHONCODE'"
	set r [TkseshDB::select $db $sql]
	return $r
    }

    # Get the list of all codes which refers to the same *glyph*
    # (Note that this excludes Phonetic codes)
    proc getEquivalentCodes {code} {
	set db [TkseshDB::getTkseshDB]

	set sql "select h2.code \
                             from HasCode as h1, HasCode as h2 \
                             where h1.code= '$code' and h1.glyphId= h2.glyphId"
	set l [TkseshDB::select $db $sql]
	return $l
    }

    proc loadBitmapFromDB {code} {
	set db [TkseshDB::getTkseshDB]
	set sql "select bitmap from SignBitmap where code = '$code'"
	set bm [TkseshDB::select $db $sql]
	return $bm
    }

    proc saveBitmapInDB {code bitmap} {
	set db [TkseshDB::getTkseshDB]
	set sql "insert into SignBitmap values ('$code', '$bitmap')"
	TkseshDB::update $db $sql
    }

    proc saveLigature {code command} {
	set db [TkseshDB::getTkseshDB] 
	set l [TkseshDB::protectText $code]
	set command [TkseshDB::protectText $command]
	if {[ligatureExists $code]} {
	    set sql "update ligature set ligRep = '$command' where ligCode = '$l'"
	} else {
	    set sql "insert into ligature values ('$l', '$command')"
	}
	TkseshDB::update $db $sql
    }

    proc forEachLigature {toApply} {
	set db [TkseshDB::getTkseshDB] 
	set sql "select ligCode, ligRep from ligature"
	TkseshDB::doSelect $db $sql tmp {
	    $toApply $tmp(ligCode) $tmp(ligRep)
	}

    } 
    
    proc ligatureExists {code} {
	set db [TkseshDB::getTkseshDB] 
	set sql "select * from ligature where ligCode = '[TkseshDB::protectText $code]'"
	return [expr {[llength [TkseshDB::select $db $sql]] !=  0}]
    }
    
    proc getLigature {code} {
	set db [TkseshDB::getTkseshDB] 
	set sql "select ligRep from ligature where ligCode = '[TkseshDB::protectText $code]'"
	return [lindex [TkseshDB::select $db $sql] 0]
    }
    
    # call "toApply" on the glyph representation for uCode
    #
    #	uCode : a glyph user code
    #   toApply : tcl code, which will be called
    #		it will receive two arguments :
    #		       * uCode
    #		       * a list representation of the glyph drawing as argument
    proc applyToGlyphByUserCode {uCode toApply} {
	set sql "\
    	      select drawing\
	      from  Glyph, HasCode\
	      where glyphId = Glyph.oid\
	      and code = '$uCode'"
	set db [TkseshDB::getTkseshDB] 
	TkseshDB::doSelect $db $sql tmp {
	    $toApply $uCode $tmp(drawing)
	}
    }

    # Find a glyph by Gardiner code, and applies "toApply" to it.
    # if "isOfficial" is false, the system will apply the method to all such glyphs.
    #
    # toApply will receive four arguments : gCode, and 
    # returns : 1 if at least a sign was found.
    proc applyToGlyphByGardinerCode {gCode toApply isOfficial} {
	set result 0
	# find the corresponding glyph code : 
	set sql \
	    "select uid, mid, id\
                 from HasCode, Header\
		 where HasCode.code = '$gCode'\
		 and HasCode.isOfficial = '$isOfficial'\
                 and glyphId = Header.repOid\
	         and Header.type = 'Glyph'"
	set db [TkseshDB::getTkseshDB] 
	TkseshDB::doSelect $db $sql tmp {

	    $toApply $gCode $tmp(uid) $tmp(mid) $tmp(id)
	    set result 1
	}
	return $result
    }

    # same system as applyToGlyphByGardinerCode.
    proc applyToGlyphByPhoneticCode {pCode toApply isOfficial} {
	set result 0
	# find the corresponding glyph code : 
	set sql \
	    "select uid, mid, id\
                 from PhonCode, HasCode, Header\
		 where PhonCode.code= '$pCode'\
		 and PhonCode.sign= HasCode.code\
		 and HasCode.isOfficial = '$isOfficial'\
                 and glyphId = Header.repOid\
	         and Header.type = 'Glyph'"
	set db [TkseshDB::getTkseshDB] 
	TkseshDB::doSelect $db $sql tmp {
	    $toApply $pCode $tmp(uid) $tmp(mid) $tmp(id)
	    set result 1
	}
	return $result
    }

    proc getVariants {code minConfidence} {
	# We look for variants of sign.
	# We consider that the "is variant"
	# relation is reflexive only if confidence is 5.
	# (signs are more or less true equivalents).
	set db [TkseshDB::getTkseshDB]
	set sql {
	    select code1 as code
	    from SignRelation 
	    where code2 = 'CODE' 
	    and confidence >= CONFIDENCE
	    UNION
	    select code2 as code
	    from SignRelation
	    where code1 = 'CODE'
	    and confidence = 5
	}
	set code [TkseshDB::protectText $code]
	set sql [string map [list CODE $code CONFIDENCE $minConfidence] $sql]
	set l [TkseshDB::select $db $sql]
	return $l
    }

    # Get the internal glyph ID for a glyph
    proc getGlyphIdFor {uid mid id} {
	set db [TkseshDB::getTkseshDB]
	set sql "select repOid from Header where type = 'Glyph' and uid = $uid and mid = $mid and id = $id"
	set l [TkseshDB::select $db $sql]
	return $l
    }
}