# STOOOP WRAPPER AROUND WIDGETS

package require BWidget 1.3
package require BWidgetBindImage
package provide myStoopWidgets 1.0
package require tkseshUtils

# we choose for classes names which are not used for BWidgets it would
# be definitly better to use a namespace here, but nested namespaces
# are really unpleasant to use : you spend your time using fully
# qualified names and so on.


# wrapper class around basic tk widgets

# TYPICAL USE :
# 
# subclass an existing class.
# the real widget creation is made by creatorHook, which looks like :
#
#     proc creatorHook {this} 
#	Component::_creatorHook $this
#       YOUR CODE HERE
#
# to create a widget, you usually want to use :
# set w [Component::buildNewComponent MyClass $parentObject -auto]

class Component {
    
    # Possibilities : we want a full OO system : 
    # TK pathes should be avoided at all cost
    #
    #	parent : the parent widget ID (NOT PATH !!!)
    #	name : the last part of the path for this widget.
    #	       if name is given as -auto, it will be computed.
    
    proc Component {this parent args} {
	# NOTE : we suppose that 0 is equivalent to NULL/null/NIL
	# we should ask Jean-Luc Fontaine
	set ($this,args) $args
	set ($this,parent) $parent
	
	# Should move to class Container
	set ($this,childNames) 0
	set ($this,hasWindow) 0 
	#puts "creating component $this"
    }
    
    proc ~Component {this} {
	# The actual widget might have been already destroyed.
	# hence this catch.
	puts "deleting component $this"

	if {$($this,hasWindow)} {
	    set ($this,hasWindow) 0
	    destroy  [getPath $this]
	}
	catch {unset (component,$path)}
    }
    

    # static method used as contructor.
    # it create a new object and build the corresponding widget
    # this is usually the one you want to call for actual creation
    proc buildNewComponent {className parent args} {
	set i [eval new $className $parent $args]
	Component::create $i
	return $i
    }


    proc getParent {this} {
	return $($this,parent)
    }

    proc cget {this var} {
	return [[getPath $this] cget -$var]
    }
    
    proc cset {this var val} {	
	[getPath $this] configure -$var $val
    }

    proc configure {this args} {	
	eval [getPath $this] configure $args
    }

    proc setArgs {this args} {
	set ($this,args) $val
    }

    proc addArgs {this args} {
	foreach i $args {
	    lappend ($this,args) $i
	}
    }

    proc getArgs {this} {
	return $($this,args)
    }

    # Component is a wrapper class. 
    # In a number of cases, we will need to wrap 
    # *existing widgets* in a Component.
    # thus, the constructor should not actually create the
    # tk widget.
    #
    # the "create" method is supposed to be called for actual widget creation,
    # if the widget does not already exist.
    
    proc create {this} {
	creatorHook $this
	complete $this
    }

    # Attach this object to an already existing TK widget.
    
    proc attach {this path} {
	setPath $this $path
	complete $this
    }

    # Subclasses must define creatorHook which takes care of actual widget creation.

    virtual proc creatorHook {this} 

    # Subclasses may define completorHook :
    # Completing is used for operations which take place AFTER actual widget creation,
    # and need the tk widget to be created.
    # This allows one to use an already existing widget, and complete it instead of
    # creating it.

    proc complete {this} {
	completorHook $this
	bind $this <Destroy> [list Component::deleteMe $this]
    }
    
    virtual proc completorHook {this} {
    }

    # Try to delete a component.
    proc deleteMe {this} {
	puts "Here deleteMe $this"
	bind $this <Destroy> {}
	if {$($this,hasWindow)} {
	    set ($this,hasWindow) 0
	    delete $this
	}
    }

    proc setCreator {className creator} {
	set (::$className,creator) $creator
    }

    proc getCreator {this} {
	return $([classof $this],creator)
    }
    # setPath is logically a protected procedure and should not be used
    # outside widget definitions

    proc setPath {this path} {
	# set this widget's path
	# and record it.
	set ($this,hasWindow) 1
	set (component,$path) $this
	set ($this,path) $path
	
    }

    virtual proc getPath {this} {
	return $($this,path)
    }

    # static procedure : for a given path, retrieve the widget's object ID.
    # usefull for binding definitions.
    
    proc getComponent {path} {
	return $(component,$path)
    }

    proc bind {this event action} {
	# set action [string map {%O {[Component::getComponent %W]}} $action]
	set action [string map [list %O $this] $action]
	::bind [getPath $this] $event $action
    }

    proc registerHelp {this helptext} {
	DynamicHelp::register [getPath $this] balloon $helptext
    }

    proc grid {subw args} {
	set l [list]
	foreach i $subw {
	    lappend l [getPath $i]
	}
	eval ::grid $l $args
    }

    proc rowConfigure {this num args} {
	eval ::grid rowconfigure [Component::getPath $this] $num $args
    }

    proc columnConfigure {this num args} {
	eval ::grid columnconfigure [Component::getPath $this] $num $args
    }

    proc pack {subw args} {
	set l [list]
	foreach i $subw {
	    lappend l [getPath $i]
	}
	eval ::pack $l $args
    }

    proc raise {this} {
	::raise [Component::getPath $this]
    }

    proc lower {this} {
	::lower [Component::getPath $this]
    }    

}

# Most components are simple components :
# they have a parent component which is an object.

class SimpleComponent {
    proc SimpleComponent {this parent name args} Component {$parent $args} {
	# name computing : 
	if {[string equal $name -auto]} {
	    set name [incr Component::($parent,childNames)]
	}
	# NOTE : we suppose that 0 is equivalent to NULL/null/NIL
	# we should ask Jean-Luc Fontaine
	if {$parent == 0} {
	    # We are in a root/toplevel widget
	    Component::setPath $this ".$name"
	} else {
	    set p [Component::getPath $parent]
	    if {[string equal $p "."]} {
		Component::setPath $this ".$name"
	    } else {
		Component::setPath $this "$p.$name"
	    }
	}
    }
    
    proc ~SimpleComponent {this} {}
    
    virtual proc creatorHook {this} {
	Component::_creatorHook $this
    }

    virtual proc completorHook {this} {
	Component::_completorHook $this
    }

}

class STopLevel {
    
    proc STopLevel {this name} SimpleComponent {0 $name} {
    }
    
    proc ~STopLevel {this} {}

    virtual proc creatorHook {this} {
	eval toplevel [Component::getPath $this] [Component::getArgs $this]
    }

    virtual proc completorHook {this} {
	Component::_completorHook $this
    }

    proc buildNewTopLevel {className name args} {
	set i [eval new $className $name $args]
	Component::create $i
	return $i
    }
}

# Standard widget definitions
foreach {w n} {
    canvas SCanvas text SText
    Label SLabel Button SButton
    radiobutton SRadioButton
    checkbutton SCheckButton
    frame SFrame
    ScrolledWindow SScrolledWindow
    LabelEntry SLabelEntry
    Entry SEntry
    ComboBox SComboBox
} {
    Component::setCreator $n $w

    class $n {
	proc $n {this parent name args} SimpleComponent {$parent $name $args}  {
	}
	
	proc ~$n {this} {
	}
	virtual proc completorHook {this} {
	}
    }

    virtual proc ${n}::creatorHook {this}  \
	    "eval $w \[Component::getPath \$this\] \[Component::getArgs \$this\]"
}

# Wrapper class around the BWidgets class

class BComponent {
    proc BComponent {this parent name} SimpleComponent {$parent $name} {
    }
    
    proc ~BComponent {this} {
	destroy [getPath $this]
    }
    
    # Class oriented methods
    
    #proc addmap {class subclass subpath options }
    #proc bwinclude class subclass subpath ?arg...? 
    #proc declare class optlist 
    #proc syncoptions class subclass subpath options 
    #proc tkinclude class tkwidget su
    
    
    # Widget oriented methods
    
    #proc cget {this option}
    #proc configure {this options}
    #proc getoption this option 
    
    #proc hasChanged this option pvalue 
    #proc setoption this option value 
    #proc subcget this subwidget 
    
    # badly documented :
    # Widget::focusNext w 
    # Widget::focusOK w 
    # Widget::focusPrev w 
    # Widget::generate-doc dir widgetlist 
    # Widget::generate-widget-doc class iscmd file 
    # Widget::init class path options 
}

# Container contain other widgets.

class SContainer {
}


# for OO NoteBooks, we only know objects, not symbolic reference.

class SNoteBook {
    proc SNoteBook {this parent name args} SimpleComponent {$parent $name $args}  {
    }
    
    proc ~SNoteBook {this} {
    }
    
    virtual proc creatorHook {this} {
	eval NoteBook [Component::getPath $this] [Component::getArgs $this]
    }

    virtual proc completorHook {this} {
    }

    proc bindtabs {this event script} {
	[Component::getPath $this] bindtabs $event $script
    }
    
    proc bindimages {this event script} {
	[Component::getPath $this] bindimages $event $script
    }

    proc compute_size {this} {	
	[Component::getPath $this] compute_size
    }
    
    # Page is a page object reference
    proc deletePage {this page {destroyframe 1}} {
	set win [Component::getPath $this]
	$win delete $page $destroyframe
	delete $page
	# ensures that a page is raised, if there is still a page :
	set l [$win pages]
	if {[llength $l] > 0} {
	    $win raise [lindex $l 0]
	}
    }


    #proc getframe {this page} {
    #	[Component::getPath $this] getframe $page
    #}

    proc index {this page} {
	[Component::getPath $this] index $page
    }

    proc insert {this index args} {
	set i [new SNoteBookPage $this]
	set p [eval [Component::getPath $this] insert $index $i $args]
	puts "The path is $p"
	Component::setPath $i $p
	return $i
    }

    #proc itemcget {this page option} {
    #}

    #proc itemconfigure {this page ?option? ?value option value ...?} {
    #}

    proc move {this page index} {
	[Component::getPath $this] index $page
    }

    proc pages {this args} {
	return [eval [Component::getPath $this] pages $args]
    }
    
    # Return the upper page, or 0 if none is available.
    
    proc getUpper {this} {
	set p [[Component::getPath $this] raise]
	puts "$p is raised"
	if [string equal $p ""] {
	    puts "returning 0"
	    return 0
	} else {
	    return $p
	}
    }
}

# A frame container is a widget which contains only one frame.

class FrameContainer {
    proc FrameContainer {this} {
	set ($this,frame) 0
    }

    proc ~FrameContainer {this} {
	if {$($this,frame)} {
	    delete $($this,frame)
	}
    }

    virtual proc getFrame {this} {
	if {$($this,frame) == 0} {
	    set p [[Component::getPath $this] getframe]
	    set f [new SFrame $this -auto]
	    set ($this,frame) $f
	    Component::setPath $f $p
	}
	return $($this,frame)
    }
}

class SLabelFrame {
    proc SLabelFrame {this parent name args} SimpleComponent {$parent $name $args}  \
	FrameContainer {} {
	}
	
    proc ~SLabelFrame {this} {
    }
    
    virtual proc completorHook {this} {
    }
    
    virtual proc creatorHook {this} {
	eval LabelFrame [Component::getPath $this] [Component::getArgs $this]
    }
}

class SNoteBookPage {
    
    proc SNoteBookPage {this parent} Component {$parent} FrameContainer {} {
    }

    proc ~SNoteBookPage {this} {
    }

    virtual proc completorHook {this} {
    }

    proc itemCGet {this var} {
	set p [Component::getParent $this]
	return [[Component::getPath $p] itemcget $this -$var]
    }

    proc itemConfigure {this args} {
	set p [Component::getParent $this]
	eval [Component::getPath $p] itemconfigure $this $args
    }
    
    proc getFrame {this} {
	set parent [Component::getParent $this]
	if {$FrameContainer::($this,frame) == 0} {
	    set p [[Component::getPath $parent] getframe $this]
	    set f [new SFrame $this -auto]
	    set ($this,frame) $f
	    Component::setPath $f $p
	}
	return $($this,frame)
    }

    proc raise {this} {
	set p [Component::getParent $this]
	[Component::getPath $p] raise $this
    }
    
    proc see {this} {
	set p [Component::getParent $this]
	[Component::getPath $p] see $this
    }

    # Sets an object associated with this page.
    # typically, a window
    proc setObject {this o} {
	set ($this,object) $o
    }

    proc getObject {this} {
	return $($this,object)
    }
}
 

class SMainFrame {
    proc SMainFrame {this parent name args} \
	SimpleComponent {$parent $name $args} \
	FrameContainer {} {
	    set ($this,nbIndicators) 0
	    set ($this,nbToolBars) 0
    }

    proc ~SMainFrame {this} {
    }

    virtual proc creatorHook {this}  {
	eval MainFrame [Component::getPath $this] [Component::getArgs $this]
    }
    
    virtual proc completorHook {this} {
    }

    proc addToolBarPath {this} {
	# Creation of a new toolbar.
	# Only the path is sent back.
	# Thus, the application can embed it in an object.
	set res [[Component::getPath $this] addtoolbar]
	return $res
    }
    
    # Creates a new toolbar.
    # arguments :
    # className : name of the class for the toolbar object
    # 	for simple stuff, it might be SFrame
    # args : args to pass to the constructor.
    # parent and name are already provided.
    #
    # return:
    #	the created toolbar object
    
    proc addToolBar {this className args} {
	set path [addToolBarPath $this]
	set tb [eval new $className $this -auto $args]
	Component::attach $tb $path
	# register the toolbar :
	set i $($this,nbToolBars)
	set ($this,toolbar,$i) $tb
	incr ($this,nbToolBars)
	
	set ($this,toolbarIdx,$tb) $i

	return $tb
    }

    proc getToolBar {this index} {
	return $($this,toolbar,$index)
    }

    #proc getMenu {this index} {
	#	[Component::getPath $this] getmenu $index
    #}

    proc setMenuState {this tag state} {
	[Component::getPath $this] setmenustate $tag $state
    }

    proc showStatusBar {this name} {
	[Component::getPath $this] $name
    }

    proc showToolBar {this tb bool} {
	[Component::getPath $this] showtoolbar $($this,toolbarIdx,$tb) $bool
    }
    
    proc addIndicator {this args} {
	eval [Component::getPath $this] addindicator $args
    }
    
    proc getIndicator {this index} {
    }
}

class SPanedWindow {

    proc SPanedWindow {this parent name args} SimpleComponent {$parent $name $args}  {
	set ($this,nbSubFrames) 0
    }
    
    proc ~SPanedWindow {this} {
	for {set i 0} {$i < [getSubFrameNumber $this]} {incr i} {
	    delete [getframe $this $i]
	}
    }

    proc  add {this args} {
	# get path
	set path [eval [Component::getPath $this] add $args]
	# get proper index for this new subframe
	set index $($this,nbSubFrames)
	# create the subframe object
	set f [new SFrame $this -auto]	
	Component::setPath $f [[Component::getPath $this] getframe $index]
	# register it
	set ($this,subframe,$index) $f
	# update the number of subframes
	incr ($this,nbSubFrames)
	# returns the created subframe :
	return $f
    }

    proc getSubFrameNumber {this} {
	return $($this,nbSubFrames)
    }

    proc getFrame {this index} {
	return $($this,subframe,$index)
    }

    virtual proc creatorHook {this}  {
	puts here
	eval PanedWindow [Component::getPath $this] [Component::getArgs $this]
    }

    virtual proc completorHook {this} {
    }
}

proc SScrolledWindow::setWidget {this sub} {
    [Component::getPath $this] setwidget [Component::getPath $sub]
    set ($this,widget) $sub
}

proc SScrolledWindow::getWidget {this} {
    return $($this,widget)
}

class SScrollableFrame {}

proc SScrollableFrame::SScrollableFrame {this parent name args} \
    SimpleComponent {$parent $name $args} \
    FrameContainer {} {
    }

proc SScrollableFrame::~SScrollableFrame {this} {
}

virtual proc SScrollableFrame::creatorHook {this} {
    eval ScrollableFrame [Component::getPath $this] [Component::getArgs $this]
}

virtual proc SScrollableFrame::completorHook {this} {
}

proc SEntry::get {this} {
    set p [Component::getPath $this]
    return [$p get]
}

proc SEntry::clear {this} {
    set p [Component::getPath $this] 
    $p delete 0 end
}

proc SEntry::setText {this txt} {
    SEntry::clear $this
    SEntry::insert $this insert $txt
}

proc SEntry::insert {this index string} {
    [Component::getPath $this]  insert $index $string
    
}

class SButtonBox {
    proc SButtonBox {this parent name args} SimpleComponent {$parent $name $args}  {
	set subbuttons []
    }
    
    proc ~SButtonBox {this} {
	foreach b $subbuttons {
	    delete $b
	}
    }
 
    proc add {this button args} {
	eval [Component::getPath $this] add [Component::getPath $button] $args
    }
    virtual proc completorHook {this} {
    }
    
    virtual proc creatorHook {this}  {
	eval ButtonBox [Component::getPath $this] [Component::getArgs $this]
    }
}

class SProgressBar {

    proc SProgressBar {this parent name args} \
	SimpleComponent {$parent $name -variable ::SProgressBar::($this,value) $args} \
	ProgressMeter {} {
	}
    
    proc ~SProgressBar {this} {
    }
    
    virtual proc creatorHook {this}  {
	eval ProgressBar [Component::getPath $this] [Component::getArgs $this]
    }

    virtual proc completorHook {this} {
    }

    virtual proc setMax {this max} {
	Component::configure $this -maximum $max
    }
    
    virtual proc notify {this value} {
	set ($this,value) $value
    }

    virtual proc notifyFinish {this} {
    }
}

class SProgressWindow {

    proc SProgressWindow {this name label} STopLevel {$name} ProgressMeter {}  {
	set ($this,label) $label
    }
    
    proc ~SProgressWindow {this} {
    }

    virtual proc completorHook {this} {

	wm title [Component::getPath $this] $($this,label)
	
	set ($this,label) [Component::buildNewComponent SLabel $this -auto \
			       -text $($this,label)]

	set ($this,bar) [Component::buildNewComponent SProgressBar $this -auto]
	
	Component::grid $($this,label)
	Component::grid $($this,bar)
    }

    virtual proc setMax {this max} {
	Component::configure $($this,bar) -maximum $max
    }
    
    virtual proc notify {this value} {
	SProgressBar::notify $($this,bar) $value
    }

    virtual proc notifyFinish {this} {
	SProgressBar::notifyFinish $this
    }
     

}
